using System;
using System.Collections.Generic;
using System.Text;
using CDBurnerXP.Burning;
using CDBurnerXP.IO;
using System.IO;
using RocketDivision.StarBurnX;
using System.Runtime.InteropServices;
using System.Threading;

namespace CDBurnerXP.CommandLine
{
    static class Commands
    {
        private static short m_Device = -1;

        public static void ListDrives()
        {
            for (short i = 0; i < BurningEngine.StarBurnEngine.Drives.Count; i++)
            {
                Drive drive = BurningEngine.StarBurnEngine.Drives[i];
                Console.WriteLine("{0:0}: {1} {2} ({3}:\\)", i, drive.DriveInfo.Vendor, drive.DriveInfo.Product, drive.DriveInfo.Letter);
            }
        }

        /// <summary>
        /// Checks if a given device ID is valid, and prints a message if it is not.
        /// </summary>
        /// <returns>false if device ID is not valid</returns>
        private static bool CheckDeviceValid(short deviceIndex)
        {
            if (deviceIndex < 0 || deviceIndex >= BurningEngine.StarBurnEngine.Drives.Count)
            {
                Console.WriteLine("Drive index '{0}' is invalid.", deviceIndex);
                return false;
            }
            m_Device = deviceIndex;
            return true;
        }

        public static void Erase(short deviceIndex)
        {
            if (!CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];
            device.DiscInfo.Refresh();

            if (device.DiscInfo.DiscType == (int)STARBURN_DISC_TYPE.STARBURN_DISC_TYPE_NONE)
            {
                Console.WriteLine("No disc loaded in drive.");
                return;
            }

            if ((device.DiscInfo.DiscType & (int)STARBURN_DISC_TYPE.STARBURN_DISC_TYPE_ERASABLE) == 0)
            {
                Console.WriteLine("Disc is not erasable.");
                return;
            }

            Console.WriteLine("Erasing disc, please wait...");
            DiscEraser eraser = new DiscEraser();
            eraser.Drive = device;
            eraser.OnProgress += new IDiscEraserEvents_OnProgressEventHandler(delegate(int Percent, int TimeRemaining)
                {
                    Console.WriteLine("{0}%", Percent);
                });
            eraser.Erase(false);

            Console.WriteLine("Finished erasing the disc.");
        }

        public static void Eject(short deviceIndex)
        {
            if (!CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];
            Console.WriteLine("Ejecting disc...");
            device.Eject();
        }

        public static void Load(short deviceIndex)
        {
            if (!CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];
            Console.WriteLine("Closing tray...");
            device.Load();
        }

        public static void BurnAudio(short deviceIndex, CommandLineOptions options)
        {
            if (!CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];

            BurningEngine.InitialiseBass();

            CDBurnerXP.Burning.AudioBurner audioBurner = new CDBurnerXP.Burning.AudioBurner();

            // Easy stuff first, set burn options
            audioBurner.Mode = options.Flags.Contains("-dao") ? STARBURN_WRITE_MODE.STARBURN_WRITE_MODE_DAO_96 : STARBURN_WRITE_MODE.STARBURN_WRITE_MODE_TAO;
            audioBurner.NextSessionAllowed = !options.Flags.Contains("-close");
            audioBurner.Drive = device;

            // Now get to the serious issues and add files
            foreach (FileFolderArgument fileOrFolder in options.FilesAndFolders)
            {
                if (fileOrFolder.IsFile)
                {
                    AddAudioFile(audioBurner, fileOrFolder.SourcePath);
                }
                else if (fileOrFolder.IsDir)
                {
                    foreach (string file in Directory.GetFiles(fileOrFolder.SourcePath))
                    {
                        AddAudioFile(audioBurner, file);
                    }
                }
            }

            audioBurner.OnProgress += new IAudioBurnerEvents_OnProgressEventHandler(delegate(int Percent, int TimeRemaining)
            {
                Console.WriteLine("{0}%", Percent);
            });

            if (audioBurner.Tracks.Count > 0)
            {
                audioBurner.WriteSpeed = GetSpeed(device, options);
                Console.WriteLine("Starting to write disc...{0} tracks", audioBurner.Tracks.Count);
                try
                {
                    audioBurner.Burn(false);
                }
                catch (COMException)
                {
                    throw new COMException(audioBurner.LastError, (int)audioBurner.LastErrorCode);
                }
                Console.WriteLine("Audio disc burned successfully.");

                if (options.Flags.Contains("-eject"))
                {
                    device.Eject();
                }
            }
            else
            {
                Console.WriteLine("No tracks to burn.");
            }
        }

        /// <summary>
        /// Adds an audio file to the audio image.
        /// If necessary, convert to wav before.
        /// </summary>
        private static void AddAudioFile(CDBurnerXP.Burning.AudioBurner audioBurner, string file)
        {
            try
            {
                if (Playlists.IsAudioFile(file))
                {
                    try
                    {
                        audioBurner.AddFile(file);
                    }
                    catch (COMException)
                    {
                        // Unsupported format. Try converting.
                        audioBurner.AddFile(AudioConverter.ToWav(file, 44100));
                    }
                    Console.WriteLine("Added: " + file);
                }
                else
                {
                    Console.WriteLine("Skipped: " + file);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        /// <summary>
        /// Determines the drive speed in KB/s based on the ratio value.
        /// </summary>
        private static int GetSpeed(Drive device, CommandLineOptions options)
        {
            if (options.Arguments.ContainsKey("-speed"))
            {
                foreach (DriveSpeed speed in device.DriveInfo.SupportedWriteSpeeds)
                {
                    if (options.Arguments["-speed"] == speed.SpeedRatio.ToString().TrimEnd('0'))
                    {
                        return speed.Speed;
                    }
                }
            }

            return 0;
        }

        public static void BurnIso(short deviceIndex, CommandLineOptions options)
        {
            if (!CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];

            if (!options.FilesAndFolders.HasAtLeastOneFile())
            {
                Console.WriteLine("No ISO image file specified.");
                return;
            }

            RocketDivision.StarBurnX.ImageBurner imgBurner = new RocketDivision.StarBurnX.ImageBurner();

            imgBurner.OnProgress += new IImageBurnerEvents_OnProgressEventHandler(delegate(int Percent, int TimeRemaining)
                {
                    Console.WriteLine("{0}%", Percent);
                });

            string sourceFile = options.FilesAndFolders.GetFirstFile().SourcePath;

            Console.WriteLine("Starting to burn ISO image from {0}...", sourceFile);

            imgBurner.Drive = device;
            imgBurner.WriteSpeed = GetSpeed(device, options);
            imgBurner.ImageFileName = sourceFile;
            imgBurner.NextSessionAllowed = !options.Flags.Contains("-close");
            imgBurner.Burn(false);

            if (options.Flags.Contains("-eject"))
            {
                device.Eject();
            }
        }

        public static void BurnData(short deviceIndex, CommandLineOptions options)
        {
            if (!options.Arguments.ContainsKey("-iso") && !CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];
            UDFDataBurner dataBurner = new UDFDataBurner();
            dataBurner.Drive = device;
            dataBurner.AutoResolveCollisions = true;
            dataBurner.FileDateTimeMode = options.Flags.Contains("-changefiledates") ? STARBURN_FILE_DATETIME_MODE.STARBURN_FILE_DATETIME_CURRENT : STARBURN_FILE_DATETIME_MODE.STARBURN_FILE_DATETIME_ORIGINAL;
            dataBurner.FileDateTimeFormat = STARBURN_DATETIME_FORMAT.STARBURN_DATETIME_FORMAT_UTC;

            dataBurner.OnProgress += new IUDFDataBurnerEvents_OnProgressEventHandler(delegate(int Percent, int TimeRemaining)
            {
                Console.WriteLine("{0}%", Percent);
            });
            dataBurner.OnVerify += new IUDFDataBurnerEvents_OnVerifyEventHandler(delegate(int Percent, int TimeRemaining)
            {
                Console.WriteLine("Verifying: {0}%", Percent);
            });

            // From layout or command line arguments?
            if (options.Arguments.ContainsKey("-layout"))
            {
                DataCompilation comp = new DataCompilation(dataBurner);
                Console.WriteLine("Loading compilation...");
                try
                {
                    comp.Load(options.Arguments["-layout"]);
                }
                catch (Exception ex)
                {
                    // Failed loading compilation, nothing to burn here
                    Console.WriteLine(ex.Message);
                    return;
                }
            }
            else
            {
                // Import if wanted
                if (device != null && options.Flags.Contains("-import"))
                {
                    try
                    {
                        dataBurner.ImportTrack(-1);
                        Console.WriteLine("Existing layout imported successfully.");
                    }
                    catch (COMException ex)
                    {
                        Console.WriteLine("Could not import existing layout ({0}).", ex.Message);
                    }
                }

                // Now get to the serious issues and add files
                foreach (FileFolderArgument fileOrFolder in options.FilesAndFolders)
                {
                    if (fileOrFolder.IsFile)
                    {
                        dataBurner.AddFileToFolder(fileOrFolder.TargetPath, fileOrFolder.SourcePath);
                    }
                    else if (fileOrFolder.IsDir)
                    {
                        dataBurner.AddFolderToFolder(fileOrFolder.TargetPath, fileOrFolder.SourcePath);
                    }
                }
            }

            // Set burn options
            dataBurner.Mode = STARBURN_WRITE_MODE.STARBURN_WRITE_MODE_AUTO;
            if (options.Flags.Contains("-dao"))
            {
                dataBurner.Mode = STARBURN_WRITE_MODE.STARBURN_WRITE_MODE_DAO_96;
            }
            else if (options.Flags.Contains("-sao"))
            {
                dataBurner.Mode = STARBURN_WRITE_MODE.STARBURN_WRITE_MODE_SAO;
            }
            else if (options.Flags.Contains("-tao"))
            {
                dataBurner.Mode = STARBURN_WRITE_MODE.STARBURN_WRITE_MODE_TAO;
            }
            if (options.Arguments.ContainsKey("-udf"))
            {
                switch (options.Arguments["-udf"])
                {
                    case "2.6": dataBurner.FileSystemVersion = STARBURN_UDF_VERSION.STARBURN_UDF_VERSION_2_60; break;
                    case "2.5": dataBurner.FileSystemVersion = STARBURN_UDF_VERSION.STARBURN_UDF_VERSION_2_50; break;
                    case "2.01": dataBurner.FileSystemVersion = STARBURN_UDF_VERSION.STARBURN_UDF_VERSION_2_01; break;
                    case "2.0": dataBurner.FileSystemVersion = STARBURN_UDF_VERSION.STARBURN_UDF_VERSION_2_00; break;
                    case "1.5": dataBurner.FileSystemVersion = STARBURN_UDF_VERSION.STARBURN_UDF_VERSION_1_50; break;
                    case "1.02": dataBurner.FileSystemVersion = STARBURN_UDF_VERSION.STARBURN_UDF_VERSION_1_02; break;
                }
            }

            dataBurner.NextSessionAllowed = !options.Flags.Contains("-close");

            // Volume type
            if (options.Arguments.ContainsKey("-format"))
            {
                Console.WriteLine("Warning: Option '-format' is no longer supported.");
            }

            string name = options.Arguments.ContainsKey("-name") ? options.Arguments["-name"] : "";

            // Burn ISO or write disc?
            if (options.Arguments.ContainsKey("-iso"))
            {
                string path = options.Arguments["-iso"].Trim();
                Console.WriteLine("Creating ISO image at {0}, {1} Files, Size: {2}", path, dataBurner.GetFileCount(), FormatFileSize.Format(dataBurner.Size * 1024L));
                dataBurner.SaveImageFile(path, name);
            }
            else
            {
                dataBurner.WriteSpeed = GetSpeed(device, options);
                Console.WriteLine("Starting to write disc...{0} Files, Size: {1}", dataBurner.GetFileCount(), FormatFileSize.Format(dataBurner.Size * 1024L));
                try
                {
                    dataBurner.Burn(false, name);
                }
                catch (COMException)
                {
                    throw new COMException(dataBurner.LastError, (int)dataBurner.LastErrorCode);
                }
                Console.WriteLine("Burning data disc completed.");

                if (options.Flags.Contains("-verify"))
                {
                    Console.WriteLine("Verifying disc...");
                    dataBurner.Verify();
                    Console.WriteLine("Verifying completed.");
                }

                if (options.Flags.Contains("-eject"))
                {
                    device.Eject();
                }
            }
        }

        public static void WaitForDisc(short deviceIndex)
        {
            if (!CheckDeviceValid(deviceIndex)) return;

            Drive device = BurningEngine.StarBurnEngine.Drives[deviceIndex];

            while (device.State != STARBURN_DRIVE_STATE.STARBURN_DRIVE_STATE_READY)
            {
                Thread.Sleep(500);
                device.DiscInfo.Refresh();
            }
        }
    }
}
