﻿using System;
using System.Collections.Generic;
using System.Text;
using CDBurnerXP.IO;
using System.ComponentModel;
using System.Reflection;
using System.Drawing;
using System.IO;

namespace CDBurnerXP
{
    /// <summary>
    /// Saves, loads, and caches application wide settings, using an INI file.
    /// </summary>
    public static class ApplicationSettings
    {
        #region IniCategory

        /// <summary>
        /// Allows to specify an INI-Category to which a
        /// property should be saved within an INI file.
        /// </summary>
        private class IniCategory : Attribute
        {
            private string category = string.Empty;
            private object defaultValue = null;

            /// <summary>
            /// Category within the INI file
            /// </summary>
            public string Category
            {
                get { return category; }
            }

            /// <summary>
            /// If no value exists in the INI file, assume this default value
            /// </summary>
            public object DefaultValue
            {
                get { return defaultValue; }
            }

            public IniCategory(string category)
            {
                this.category = category;
            }

            public IniCategory(string category, object defaultValue)
            {
                this.category = category;
                this.defaultValue = defaultValue;
            }
        }

        #endregion

        #region Properties

        [IniCategory("General")]
        public static string AppLanguage { get; set; }

        [IniCategory("General")]
        public static bool ExitToNotify { get; set; }

        [IniCategory("General", 2)]
        public static int ToolbarDisplayStyle { get; set; }

        [IniCategory("General")]
        public static bool OnlyUseSingleUnit { get; set; }

        [IniCategory("General")]
        public static bool StartLastUsedProjectType { get; set; }

        [IniCategory("General")]
        public static bool LoggingEnabled { get; set; }

        [IniCategory("General")]
        public static bool UseSPTD { get; set; }

        [IniCategory("General")]
        public static bool UseCustomTempFolder { get; set; }

        [IniCategory("General")]
        public static int LastUsedProjectType { get; set; }

        [IniCategory("General")]
        public static string CustomTempFolder { get; set; }

        [IniCategory("General", true)]
        public static bool AutoUpdate { get; set; }
        
        [IniCategory("General", false)]
        public static bool PreventDefaultDragDrop { get; set; }

        [IniCategory("Data")]
        public static string DataExplorerPath { get; set; }
        
        [IniCategory("Data", true)]
        public static bool DataExplorer { get; set; }

        [IniCategory("Data", false)]
        public static bool IgnoreFileRenameRequirements { get; set; }

        [IniCategory("Data", true)]
        public static bool DataCacheFilesAutomatically { get; set; }

        [IniCategory("Data", false)]
        public static bool DataCacheNetworkFilesAutomatically { get; set; }

        [IniCategory("Data")]
        public static bool DataUpdateAutomatically { get; set; }
        
        [IniCategory("Data")]
        public static bool DataBurnVerify { get; set; }

        [IniCategory("Data")]
        public static bool DataBurnFinalize { get; set; }

        [IniCategory("Data")]
        public static bool DataUseDiscAtOnce { get; set; }
        
        [IniCategory("Data", true)]
        public static bool DataBurnEject { get; set; }
        
        [IniCategory("Data")]
        public static bool DataTileVertical { get; set; }

        [IniCategory("Data")]
        public static bool DataMode2XA { get; set; }

        [IniCategory("Data", true)]
        public static bool DataAskForSettings { get; set; }

        [IniCategory("Data")]
        public static Point DataDropFormLocation { get; set; }

        [IniCategory("Data", "")]
        public static string DataFilePattern { get; set; }

        [IniCategory("Data", true)]
        public static bool DataFilePatternInclude { get; set; }

        [IniCategory("Data", true)]
        public static bool DataFilePatternShowInBrowser { get; set; }

        [IniCategory("Data")]
        public static string DataSaveISOPath { get; set; }

        [IniCategory("Data")]
        public static string DataVideoLayoutPath { get; set; }

        [IniCategory("Data", 1)]
        public static int DataExplorerViewStyle { get; set; }

        [IniCategory("Data")]
        public static int DataSplitterExplorer { get; set; }

        [IniCategory("Data")]
        public static int DataSplitterVertical { get; set; }

        [IniCategory("Data")]
        public static int DataSplitterLayout { get; set; }

        [IniCategory("Data")]
        public static long LastDataBurnSpeed { get; set; }

        [IniCategory("Data", -1)]
        public static int DataFileSystem { get; set; }

        [IniCategory("Data", -1)]
        public static int DataIsoLevel { get; set; }

        [IniCategory("Data")]
        public static string DataLastDeviceUsed { get; set; }

        [IniCategory("Data")]
        public static string DataFloppyImagePath { get; set; }
        
        [IniCategory("Data")]
        public static string DataCopyImagePath { get; set; }

        [IniCategory("Data")]
        public static string CopySourceDevice { get; set; }

        [IniCategory("Data")]
        public static bool CopyDisableHardwareEC { get; set; }

        [IniCategory("Data")]
        public static bool CopyIgnoreUnreadableData { get; set; }

        [IniCategory("Data", 3)]
        public static int CopyReadRetries { get; set; }

        [IniCategory("Data")]
        public static int CopyImageType { get; set; }

        [IniCategory("Data")]
        public static bool CopyEjectDisc { get; set; }

        [IniCategory("Data", true)]
        public static bool CopyFinaliseDisc { get; set; }

        [IniCategory("Data")]
        public static string DataConvertTargetPath { get; set; }

        [IniCategory("Data")]
        public static string DataSaveCompilationPath { get; set; }

        [IniCategory("Data", "Disc")]
        public static string DataLabel { get; set; }

        [IniCategory("Data")]
        public static string DataCachePath { get; set; }

        [IniCategory("Data", true)]
        public static bool DoubleLayerCompatible { get; set; }

        [IniCategory("Data", true)]
        public static bool PadDvdSize { get; set; }

        [IniCategory("Data", true)]
        public static bool ModifyIfoBupFiles { get; set; }

        [IniCategory("Data")]
        public static string VideoLastDeviceUsed { get; set; }

        [IniCategory("Iso", true)]
        public static bool IsoEjectDisc { get; set; }

        [IniCategory("Iso", true)]
        public static bool IsoFinalize { get; set; }

        [IniCategory("Iso")]
        public static bool IsoMode2XA { get; set; }

        [IniCategory("Iso")]
        public static string IsoLastDeviceUsed { get; set; }

        [IniCategory("Iso")]
        public static string LastUsedIsoPath { get; set; }

        [IniCategory("Audio")]
        public static string AudioExplorerPath { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioAutoCDText { get; set; }

        [IniCategory("Audio", 1)]
        public static int AudioExplorerViewStyle { get; set; }

        [IniCategory("Audio", 0)]
        public static int AudioReplayGain { get; set; }

        [IniCategory("Audio")]
        public static bool AudioTileVertical { get; set; }

        [IniCategory("Audio")]
        public static int AudioSplitterExplorer { get; set; }

        [IniCategory("Audio", 2)]
        public static int AudioPauseLength { get; set; }

        [IniCategory("Audio")]
        public static int AudioSplitterVertical { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioExplorer { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioCloseDisc { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioBurnEject { get; set; }

        [IniCategory("Audio")]
        public static bool AudioBurnGapless { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioShowPlayer { get; set; }

        [IniCategory("Audio")]
        public static Size AudioPlayerSize { get; set; }

        [IniCategory("Audio")]
        public static Point AudioPlayerPos { get; set; }

        [IniCategory("Audio")]
        public static string AudioSaveCompilationPath { get; set; }

        [IniCategory("Audio")]
        public static string AudioLastDeviceUsed { get; set; }

        [IniCategory("Audio", "")]
        public static string AudioFilePattern { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioFilePatternInclude { get; set; }

        [IniCategory("Audio", true)]
        public static bool AudioFilePatternShowInBrowser { get; set; }

        [IniCategory("Audio")]
        public static long LastAudioBurnSpeed { get; set; }

        [IniCategory("Audio", true)]
        public static bool CDTextEnabled { get; set; }

        #endregion

        /// <summary>
        /// Specifies the INI file which is to use for saving user settings.
        /// </summary>
        /// <returns></returns>
        private static IniFile GetIniFile()
        {
            if (IsPortable())
            {
                string startupPath = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location);
                // Save settings in different file, otherwise they will be 
                // overwritten by new portable versions
                return new IniFile(Path.Combine(startupPath, "UserSettings.ini"));
            }

            return new IniFile("UserSettings", true);
        }

        /// <summary>
        /// Determines whether or not a portable version is being used.
        /// </summary>
        public static bool IsPortable()
        {
            string startupPath = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location);
            IniFile portableConfig = new IniFile(Path.Combine(startupPath, "Config.ini"));
            return portableConfig.GetBoolValue("CDBurnerXP", "Portable", false);
        }

        /// <summary>
        /// Loads all values from the INI file to memory.
        /// </summary>
        public static void Load()
        {
            IniFile file = GetIniFile();

            foreach (PropertyInfo property in typeof(ApplicationSettings).GetProperties(BindingFlags.Public | BindingFlags.Static))
            {
                // Determine category
                IniCategory categoryAttribute = Attribute.GetCustomAttribute(property, typeof(IniCategory)) as IniCategory;
                string iniCategory = (categoryAttribute == null) ? "General" : categoryAttribute.Category;

                // Get value from INI file
                if (property.PropertyType == typeof(string))
                {
                    property.SetValue(null, file.GetValue(iniCategory, property.Name, categoryAttribute.DefaultValue as string), null);
                }
                else if (property.PropertyType == typeof(bool))
                {
                    property.SetValue(null, file.GetBoolValue(iniCategory, property.Name, Conversion.ToBoolean(categoryAttribute.DefaultValue)), null);
                }
                else if (property.PropertyType == typeof(int))
                {
                    string defaultInt = (categoryAttribute.DefaultValue == null) ? "0" : categoryAttribute.DefaultValue.ToString();
                    property.SetValue(null, Conversion.ToInt(file.GetValue(iniCategory, property.Name, defaultInt)), null);
                }
                else if (property.PropertyType == typeof(long))
                {
                    string defaultInt = (categoryAttribute.DefaultValue == null) ? "0" : categoryAttribute.DefaultValue.ToString();
                    property.SetValue(null, Conversion.ToLong(file.GetValue(iniCategory, property.Name, defaultInt)), null);
                }
                else if (property.PropertyType == typeof(Point))
                {
                    property.SetValue(null, file.GetPoint(iniCategory, property.Name), null);
                }
                else if (property.PropertyType == typeof(Size))
                {
                    property.SetValue(null, file.GetSize(iniCategory, property.Name), null);
                }
            }
        }

        /// <summary>
        /// Saves all values of all properties to the INI file.
        /// </summary>
        public static void Save()
        {
            IniFile file = GetIniFile();

            foreach (PropertyInfo property in typeof(ApplicationSettings).GetProperties(BindingFlags.Public | BindingFlags.Static))
            {
                // Determine category
                IniCategory categoryAttribute = Attribute.GetCustomAttribute(property, typeof(IniCategory)) as IniCategory;
                string iniCategory = (categoryAttribute == null) ? "General" : categoryAttribute.Category;

                // Get value from INI file
                object value = property.GetValue(null, null);
                if (property.PropertyType == typeof(string))
                {
                    file.SetValue(iniCategory, property.Name, value as string);
                }
                else if (property.PropertyType == typeof(bool))
                {
                    file.SetBoolValue(iniCategory, property.Name, (bool)value);
                }
                else if (property.PropertyType == typeof(int))
                {
                    file.SetValue(iniCategory, property.Name, ((int)value).ToString());
                }
                else if (property.PropertyType == typeof(long))
                {
                    file.SetValue(iniCategory, property.Name, ((long)value).ToString());
                }
                else if (property.PropertyType == typeof(Point))
                {
                    file.SetPoint(iniCategory, property.Name, (Point)value);
                }
                else if (property.PropertyType == typeof(Size))
                {
                    file.SetSize(iniCategory, property.Name, (Size)value);
                }
            }
        }
    }
}
