﻿using System;
using System.Collections.Generic;
using System.Text;
using RocketDivision.StarBurnX;
using System.IO;
using System.Runtime.InteropServices;

namespace System.Runtime.CompilerServices
{
    public sealed class ExtensionAttribute : Attribute { }
}

namespace CDBurnerXP.Burning
{
    public static class COMExceptionExtensions
    {
        /// <summary>
        /// Determines if the given COM exception matches a particular StarBurn error code.
        /// </summary>
        public static bool IsError(this COMException ex, STARBURN_RESULT_CODES error)
        {
            return (ex.ErrorCode & (int)error) == (int)error;
        }
    }

    public static class SessionExtensions
    {
        public static long GetSize(this ISession session)
        {
            long size = 0;
            foreach (ITrack track in session.Tracks)
            {
                size += track.LogicalBlockSize * (long)track.Size;
            }

            return size;
        }
    }

    public static class DiscInfoExtensions
    {
        /// <summary>
        /// Returns the size of a given session (1-x) in bytes.
        /// </summary>
        public static long GetSessionSize(this DiscInfo discInfo, int sessionNumber)
        {
            ISession session = null;

            if (discInfo.Sessions.Count == 0) return 0;

            if (sessionNumber <= 0)
            {
                session = discInfo.Sessions[discInfo.Sessions.Count - 1];
            }
            else if (sessionNumber < discInfo.Sessions.Count)
            {
                session = discInfo.Sessions[sessionNumber - 1];
            }
            else
            {
                return 0;
            }

            return session.GetSize();
        }

        /// <summary>
        /// Returns the total used size of a disc (going through all sessions and discs).
        /// </summary>
        public static long GetTotalUsedSize(this DiscInfo discInfo)
        {
            long totalUsedSpace = 0;

            foreach (Session discSession in discInfo.Sessions)
            {
                foreach (Track sessionTrack in discSession.Tracks)
                {
                    totalUsedSpace += ((long)sessionTrack.Size) * sessionTrack.LogicalBlockSize;
                }
            }

            return totalUsedSpace;
        }

        /// <summary>
        /// Returns a TimeSpan which equals the length of audio data on the disc.
        /// </summary>
        /// <param name="usedSpaceBlocks">Number of used blocks on the disc</param>
        public static TimeSpan GetAudioDuration(this DiscInfo discInfo, int logicalBlocks)
        {
            // SecSize = (LBSizeValue / 75) 
            return TimeSpan.FromSeconds(logicalBlocks / 75.0);
        }

        /// <summary>
        /// Specifies whether or not the disc is of type CD.
        /// </summary>
        public static bool IsCDMediaType(this DiscInfo discInfo)
        {
            return (discInfo.DiscType & (int)STARBURN_DISC_TYPE.STARBURN_DISC_TYPE_CD) > 0;
        }
    }

    /// <summary>
    /// Fixes some shortcomings of the DataBurner class.
    /// </summary>
    public static class DataBurnerExtensions
    {
        /// <summary>
        /// Creates a folder on the disc if it does not yet exist and
        /// returns a reference to the folder on the disc.
        /// </summary>
        private static IDataFolder CreateFolder(UDFDataBurner dataBurner, string targetFolder)
        {
            string[] parts = targetFolder.TrimStart('\\').Split('\\');
            DataFolder existingFolder = null;
            DataFolder createNextFolderIn = (DataFolder)dataBurner;
            foreach (string part in parts)
            {
                if (string.IsNullOrEmpty(part)) continue;

                existingFolder = createNextFolderIn.GetFileObject(part + "\\") as DataFolder;
                if (existingFolder == null)
                {
                    existingFolder = createNextFolderIn.CreateFolder(part);
                }
                createNextFolderIn = existingFolder;
            }
            return existingFolder ?? (IDataFolder)dataBurner;
        }

        public static void AddFileToFolder(this UDFDataBurner dataBurner, string folderOnDisc, string filename)
        {
            string nameOnMedia = Path.GetFileName(filename);
            IDataFolder existingFolder = CreateFolder(dataBurner, folderOnDisc);
            AddFileToFolder(existingFolder, filename, nameOnMedia);
            
        }

        /// <summary>
        /// Adds or replaces a file in a folder.
        /// </summary>
        /// <param name="existingFolder">Folder to add the file to</param>
        /// <param name="filename">Source file name</param>
        /// <param name="nameOnMedia">Name of file on disc</param>
        private static void AddFileToFolder(IDataFolder targetFolder, string filename, string nameOnMedia)
        {
            DataFile existing = targetFolder.GetFileObject(nameOnMedia);
            if (existing != null)
            {
                targetFolder.Files.Remove(existing);
            }
            targetFolder.AddFileEx(filename, nameOnMedia);
        }

        public static void AddFolderToFolder(this UDFDataBurner dataBurner, string folderOnDisc, string path)
        {
            // Assume:
            // folderOnDisc:    \x
            // path:            c:\temp
            // All contents of temp (subfolders) then go into \x on the disc.

            IDataFolder existingFolder = CreateFolder(dataBurner, folderOnDisc);
            foreach (string file in Directory.GetFiles(path))
            {
                AddFileToFolder(existingFolder, file, Path.GetFileName(file));
            }
            foreach (string dir in Directory.GetDirectories(path))
            {
                dataBurner.AddFolderToFolder(folderOnDisc + "\\" + Path.GetFileName(dir), dir);
            }
        }

        /// <summary>
        /// Returns the total file count of the compilation.
        /// </summary>
        public static int GetFileCount(this UDFDataBurner dataBurner)
        {
            return CountFiles(dataBurner.Files);
        }

        private static int CountFiles(DataFiles dataFiles)
        {
            int count = 0;
            foreach (DataFile file in dataFiles)
            {
                DataFolder folder = file as DataFolder;
                if (folder != null)
                {
                    count += CountFiles(folder.Files);
                }
                else
                {
                    count++;
                }
            }

            return count;
        }
    }
}
