﻿using System;
using System.Collections.Generic;
using System.Text;
using CDBurnerXP.Controls;
using NMSDVDXLib;
using System.ComponentModel;
using System.Threading;
using CDBurnerXP.IO;
using System.IO;

namespace CDBurnerXP.Burning
{
    /// <summary>
    /// Spreads data on multiple discs.
    /// </summary>
    public class DiscSplitter
    {
        #region SplitProgressChangedEventArgs
        
        public class SplitProgressChangedEventArgs : EventArgs
        {
            private int progress;
            private int currentNumberOfDiscs;

            public int CurrentNumberOfDiscs
            {
                get { return currentNumberOfDiscs; }
            }

            public int Progress
            {
                get { return progress; }
            }

            public SplitProgressChangedEventArgs(int progress, int currentNumberOfDiscs)
            {
                this.progress = progress;
                this.currentNumberOfDiscs = currentNumberOfDiscs;
            }
        }

        #endregion

        #region SplitCompletedEventArgs

        public class SplitCompletedEventArgs : EventArgs
        {
            private List<IDataImageX> dataImages;
            private FileNode fileTooBig;

            public List<IDataImageX> DataImages
            {
                get { return dataImages; }
            }

            internal FileNode FileTooBig
            {
                get { return fileTooBig; }
            }

            internal SplitCompletedEventArgs(List<IDataImageX> dataImages, FileNode fileTooBig)
            {
                this.dataImages = dataImages;
                this.fileTooBig = fileTooBig;
            }
        }

        #endregion

        public enum Algorithm
        {
            OptimiseDiscCount,
            OptimiseFolderStructure
        }

        private long maxDiscSize = 0;
        private bool cancel = false;
        private bool busy = false;
        private Algorithm algorithm = Algorithm.OptimiseDiscCount;
        private bool tableOfContents = false;
        private string tableOfContentsName = "Table of contents";

        public event EventHandler<SplitProgressChangedEventArgs> ProgressChanged;
        public event EventHandler<SplitCompletedEventArgs> Completed;

        #region Properties

        /// <summary>
        /// Gets or sets the TOC file name (without extension).
        /// </summary>
        public string TableOfContentsName
        {
            get { return tableOfContentsName; }
            set { tableOfContentsName = value; }
        }


        /// <summary>
        /// Gets or sets whether to include a table of contents (as plain text file) to each disc.
        /// </summary>
        public bool TableOfContents
        {
            get { return tableOfContents; }
            set { tableOfContents = value; }
        }

        /// <summary>
        /// Gets or sets the method that is used to split the discs.
        /// </summary>
        public Algorithm SplitAlgorithm
        {
            get { return algorithm; }
            set { algorithm = value; }
        }

        /// <summary>
        /// Gets whether or not the working thread is busy.
        /// </summary>
        public bool IsBusy
        {
            get { return this.busy; }
        }

        /// <summary>
        /// Gets whether or not the process has been cancelled.
        /// </summary>
        public bool Cancelled
        {
            get
            {
                return this.cancel;
            }
        }

        /// <summary>
        /// Gets or sets the maximum size per data image.
        /// </summary>
        public long MaxDiscSize
        {
            get { return maxDiscSize; }
            set { maxDiscSize = value; }
        }

        #endregion

        /// <summary>
        /// Generates multiple data images with all data of the original
        /// data image divided, so that each image is filled optimally.
        /// </summary>
        public void BeginSplitLayout(DataCompilation compilation)
        {
            Thread thread = new Thread(new ParameterizedThreadStart(SplitLayout));
            thread.Start(compilation);
        }

        /// <summary>
        /// Cancels the dividing process.
        /// </summary>
        public void Cancel()
        {
            this.cancel = true;
        }

        private void SplitLayout(object arg)
        {
            this.busy = true;
            this.cancel = false;
            DataCompilation compilation = arg as DataCompilation;
            List<FileNode> files = new List<FileNode>();
            AddFilesRecursive(files, compilation.Layout);

            long maxAvailableSize = this.maxDiscSize;

            // Estimate table of contents size
            if (this.tableOfContents)
            {
                StringBuilder sb = new StringBuilder();

                foreach (FileNode file in files)
                {
                    sb.AppendLine(file.FullVirtualPath);
                }

                // Reduce available disc size + buffer for "Disc n" indicators
                maxAvailableSize -= Encoding.UTF8.GetByteCount(sb.ToString()) + 2000;
            }
            // Sometimes it does not work properly for DVDs due to insufficient disc space,
            // reduce a little more (1MB)
            if (maxAvailableSize >= DiscType.DVDPlus.SizeInBytes)
            {
                maxAvailableSize -= 1024 * 1024;
            }

            foreach (FileNode file in files)
            {
                // Check greatest file: If too huge for single disc, splitting is not possible.
                if (file.Size > maxAvailableSize)
                {
                    this.Cancel();
                    OnCompleted(null, file);
                    return;
                }
            }

            if (this.algorithm == Algorithm.OptimiseDiscCount)
            {
                // First Fit Decreasing
                files.Sort(delegate(FileNode x, FileNode y)
                {
                    return x.Size.CompareTo(y.Size);
                });
            }
            else
            {
                // Sort by folder, reverse, since we need to go through the list backwards
                files.Sort(delegate(FileNode x, FileNode y)
                {
                    return y.FullVirtualPath.CompareTo(x.FullVirtualPath);
                });
            }

            List<IDataImageX> dataImages = new List<IDataImageX>();
            List<StringBuilder> tableOfContents = new List<StringBuilder>();

            int totalFiles = files.Count;

            while (files.Count > 0)
            {
                IDataImageX newImage = BurningEngine.NmsEngine.CreateDataImage();
                newImage.ISOLevel = compilation.DataImage.ISOLevel;
                newImage.VolumeType = compilation.DataImage.VolumeType;
                newImage.UseLongJolietNames = compilation.DataImage.UseLongJolietNames;
                newImage.GetVolumeInfo().VolumeID = compilation.DataImage.GetVolumeInfo().VolumeID;
                dataImages.Add(newImage);
                StringBuilder toc = new StringBuilder();
                tableOfContents.Add(toc);

                // Start allocating the biggest files
                for (int i = files.Count - 1; i >= 0; i--)
                {
                    if (this.cancel)
                    {
                        OnCompleted(dataImages, null);
                        return;
                    }

                    newImage.GetFileLayout().AddItem(files[i].FileSystemPath, files[i].FullVirtualPath);
                    // Disc full? Start next image.
                    // We have to add first and then remove as necessary later, because each file
                    // increases the image size by a bit more than the actual file size.
                    if (newImage.GetImageSizeBytes() > maxAvailableSize)
                    {
                        newImage.GetFileLayout().RemoveItem(files[i].FullVirtualPath);
                        break;
                    }

                    long actualSize = (long)newImage.GetImageSizeBytes();
                    toc.AppendLine(files[i].FullVirtualPath);
                    files.RemoveAt(i);
                    OnProgressChanged((totalFiles - files.Count) / (float)totalFiles * 100.0f, dataImages.Count);
                }
            }

            // Generate TOC and add to all discs
            if (this.tableOfContents)
            {
                StringBuilder fullToc = new StringBuilder();

                for (int i = 0; i < tableOfContents.Count; i++)
                {
                    fullToc.AppendLine("====== Disc " + (i + 1) + " ======");
                    fullToc.AppendLine(tableOfContents[i].ToString());
                }

                try
                {
                    string tmpFile = PathEx.GetTempFileName(".txt");
                    File.WriteAllText(tmpFile, fullToc.ToString(), Encoding.UTF8);

                    for (int i = 0; i < dataImages.Count; i++)
                    {
                        dataImages[i].GetFileLayout().AddItem(tmpFile, this.tableOfContentsName + ".txt");
                    }
                }
                catch (Exception ex)
                {
                    // Do not fail the whole process for a TOC
                    Debug.LogError(ex);
                }
            }

            // Adjust disc names, add (1) to (X)
            if (compilation.Name.Length <= 28)
            {
                for (int i = 0; i < dataImages.Count; i++)
                {
                    dataImages[i].GetVolumeInfo().VolumeID = compilation.Name + " (" + (i + 1) + ")";
                }
            }

            OnCompleted(dataImages, null);
        }

        private void OnCompleted(List<IDataImageX> dataImages, FileNode fileTooBig)
        {
            this.busy = false;

            if (Completed != null)
            {
                Completed(this, new SplitCompletedEventArgs(dataImages, fileTooBig));
            }
        }

        private void OnProgressChanged(float progress, int discCount)
        {
            if (ProgressChanged != null)
            {
                ProgressChanged(this, new SplitProgressChangedEventArgs((int)progress, discCount));
            }
        }

        private void AddFilesRecursive(List<FileNode> files, DirectoryNode root)
        {
            foreach (FileNode node in root.Files.Values)
            {
                files.Add(node as FileNode);
            }
            foreach (DirectoryNode node in root.Directories.Values)
            {
                AddFilesRecursive(files, node);
            }
        }
    }
}
