using System;
using System.Collections.Generic;
using System.Text;
using NMSDVDXLib;
using RocketDivision.StarBurnX;
using CDBurnerXP.Burning;
using System.Windows.Forms;
using System.ComponentModel;
using CDBurnerXP.IO;
using System.Runtime.InteropServices;
using Microsoft.Win32;

namespace CDBurnerXP.Controls
{
    /// <summary>
    /// Represents a special ComoBox, which allows to choose
    /// from an installed optical device.
    /// </summary>
    public class DeviceSelect : ComboBox
    {
        #region DriveItem

        private class DriveItem
        {
            private string m_DeviceLetter = "";
            private string m_Text = string.Empty;
            private Drive cachedDrive = null;

            public DeviceX Device
            {
                get
                {
                    return BurningEngine.GetNmsDriveByLetter(m_DeviceLetter);
                }
            }

            public Drive StarBurnDevice
            {
                get
                {
                    if (this.cachedDrive == null)
                    {
                        this.cachedDrive = BurningEngine.GetStarBurnDriveByLetter(m_DeviceLetter);
                    }
                    return this.cachedDrive;
                }
            }

            public string DeviceLetter
            {
                get
                {
                    return m_DeviceLetter;
                }
            }

            public DriveItem(IDeviceX device)
            {
                m_DeviceLetter = device.GetDriveLetter();
                m_Text = BurningEngine.FormatDevice(device);
            }

            public DriveItem(Drive device)
            {
                m_DeviceLetter = device.DriveInfo.Letter;
                m_Text = BurningEngine.FormatDevice(device);
            }

            public override string ToString()
            {
                return m_Text;
            }
        }

        #endregion

        private bool m_IncludeReadonly = false;
        private bool m_HasDrives = false;
        private bool m_UseStarBurn = false;
        private static string m_TranslationNoDrives = "No compatible drives found.";

        #region Properties

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public new bool Enabled
        {
            get
            {
                return base.Enabled;
            }
            set
            {
                base.Enabled = value;
            }
        }

        /// <summary>
        /// Gets or sets the translated string for "No compatible drives".
        /// </summary>
        [DefaultValue("No compatible drives found.")]
        public static string TranslationNoDrives
        {
            get
            {
                return m_TranslationNoDrives;
            }
            set {
                m_TranslationNoDrives = value;
            }
        }

        /// <summary>
        /// Gets whether or not any drive is available.
        /// </summary>
        [Browsable(false)]
        public bool HasDrives
        {
            get
            {
                return m_HasDrives;
            }
        }

        /// <summary>
        /// Gets or sets whether non-burning devices should
        /// be included to the selection.
        /// </summary>
        [DefaultValue(false)]
        public bool IncludeReadonly
        {
            get
            {
                return m_IncludeReadonly;
            }
            set
            {
                if (value != m_IncludeReadonly)
                {
                    m_IncludeReadonly = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the currently selected NMS device.
        /// </summary>
        [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public DeviceX SelectedDevice
        {
            get
            {
                DriveItem item = SelectedItem as DriveItem;
                if (item == null) return null;

                return item.Device;
            }
            set
            {
                if (value == null) return;
                SelectedDeviceLetter = value.GetDriveLetter();
            }
        }

        /// <summary>
        /// Gets or sets the currently selected StarBurn device.
        /// </summary>
        [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public Drive SelectedStarBurnDevice
        {
            get
            {
                DriveItem item = SelectedItem as DriveItem;
                if (item == null) return null;

                return item.StarBurnDevice;
            }
            set
            {
                if (value == null) return;
                SelectedDeviceLetter = value.DriveInfo.Letter;
            }
        }

        /// <summary>
        /// Gets or sets the currently selected device by drive letter.
        /// </summary>
        [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public string SelectedDeviceLetter
        {
            get
            {
                DriveItem item = SelectedItem as DriveItem;
                if (item == null) return null;

                return item.DeviceLetter;
            }
            set
            {
                if (this.Items.Count == 0)
                {
                    LoadDrives();
                }

                if (!HasDrives && this.Items.Count > 0)
                {
                    // We cannot set anything but the first item
                    SelectedIndex = 0;
                    return;
                }

                foreach (DriveItem device in Items)
                {
                    if (string.Compare(device.DeviceLetter, value, true) == 0)
                    {
                        SelectedItem = device;
                        return;
                    }
                }

                SelectedIndex = 0;
            }
        }

        /// <summary>
        /// Gets or sets whether or not the StarBurn
        /// burning engine should be used.
        /// </summary>
        [DefaultValue(false)]
        public bool UseStarBurn
        {
            get { return m_UseStarBurn; }
            set 
            {
                if (m_UseStarBurn != value)
                {
                    m_UseStarBurn = value;
                }
            }
        }

        #endregion

        public DeviceSelect()
        {
            DropDownStyle = ComboBoxStyle.DropDownList;
            
        }

        protected override void OnVisibleChanged(EventArgs e)
        {
            if (this.Items.Count == 0 && this.Visible)
            {
                LoadDrives();

                if (this.SelectedIndex == -1 && this.Items.Count > 0)
                {
                    this.SelectedIndex = 0;
                }
            }

            base.OnVisibleChanged(e);
        }

        private void LoadDrives()
        {
            if (DesignMode) return;

            Items.Clear();

            if (m_UseStarBurn)
            {
                for (int i = 0; i < BurningEngine.StarBurnEngine.Drives.Count; i++)
                {
                    try
                    {
                        Drive drive = BurningEngine.StarBurnEngine.Drives[i];
                        if ((IncludeReadonly || drive.DriveInfo.SupportedMediaWrite > 0) && !PathEx.IsDriveHidden(drive.DriveInfo.Letter[0]))
                        {
                            Items.Add(new DriveItem(drive));
                        }
                    }
                    catch (COMException ex)
                    {
                        if ((ex.ErrorCode & 0xFFF) == (int)STARBURN_RESULT_CODES.STARBURN_ERR_DEVICE_DISCONNECTED)
                        {
                            // Drive does no longer exist. Do not show.
                            continue;
                        }

                        Debug.LogError(ex);

                        throw;
                    }
                }
            }
            else
            {
                for (short i = 0; i < BurningEngine.NmsEngine.GetDeviceCount(); i++)
                {
                    DeviceX device = BurningEngine.NmsEngine.GetDevice(i);
                    if ((IncludeReadonly || device.IsWriter()) && !PathEx.IsDriveHidden(device.GetDriveLetter()[0]))
                    {
                        Items.Add(new DriveItem(device));
                    }
                }
            }

            if (Items.Count <= 0)
            {
                Items.Add(m_TranslationNoDrives);
                Enabled = m_HasDrives = false;                
            }
            else
            {
                Enabled = m_HasDrives = true;
            }
        }

        protected override void OnEnabledChanged(EventArgs e)
        {
            if (!HasDrives)
            {
                Enabled = false;
            }
            else
            {
                base.OnEnabledChanged(e);
            }
        }
    }
}
