using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using NMSDVDXLib;
using Microsoft.Win32;

namespace CDBurnerXP.Controls
{
    internal class DirectoryNode : IFileSystemNode
    {
        #region FileDictionary

        internal class FileDictionary : Dictionary<string, FileNode>
        {
            private long m_SumFileSize = 0;

            public long SumFileSize
            {
                get
                {
                    return m_SumFileSize;
                }
            }

            public new void Clear()
            {
                base.Clear();
                m_SumFileSize = 0;
            }

            public new FileNode this[string key]
            {
                get
                {
                    return base[key];
                }
                set
                {
                    if (base.ContainsKey(key))
                    {
                        m_SumFileSize -= base[key].Size;
                    }
                    base[key] = value;
                    m_SumFileSize += value.Size;
                }
            }

            public new void Add(string key, FileNode value)
            {
                base.Add(key, value);
                m_SumFileSize += value.Size;
            }

            public new void Remove(string key)
            {
                m_SumFileSize -= base[key].Size;
                base.Remove(key);
            }
        }

        #endregion

        private Dictionary<string, DirectoryNode> m_SubDirs = new Dictionary<string, DirectoryNode>();
        private string m_FileSystemPath = null;
        private DirectoryNode m_Parent = null;
        private string m_Name = null;
        private FileLayoutManager m_Manager = null;
        private bool m_IsImported = false;
        private bool m_VirtualDirectory = false;
        private FileDictionary m_Files = new FileDictionary();
        private bool m_Hidden = false;
        private DateTime m_ImportedFileDate = DateTime.MinValue;

        #region Properties

        public DateTime ModificationDate
        {
            set
            {
                m_ImportedFileDate = value;
            }
            get
            {
                if (m_ImportedFileDate != DateTime.MinValue)
                {
                    return m_ImportedFileDate;
                }
                return DateTime.MinValue;
            }
        }

        public bool Hidden
        {
            get
            {
                return m_Hidden;
            }
            set
            {
                m_Hidden = value;
            }
        }

        internal FileLayoutManager Manager
        {
            get
            {
                if (m_Manager == null && m_Parent != null)
                {
                    return m_Parent.Manager;
                }
                return m_Manager;
            }
        }

        /// <summary>
        /// Gets and sets whether or not this node is virtual.
        /// A node is virtual if it has only been created for adding
        /// a new folder.
        /// </summary>
        public bool VirtualDirectory
        {
            get
            {
                return m_VirtualDirectory;
            }
            set
            {
                m_VirtualDirectory = value;
            }
        }

        public DirectoryNode Parent
        {
            get
            {
                return m_Parent;
            }
            set
            {
                m_Parent = value;
            }
        }

        public bool IsImported
        {
            get
            {
                return m_IsImported;
            }
            set
            {
                m_IsImported = value;
            }
        }

        /// <summary>
        /// Determines the size of all contents.
        /// </summary>
        public long Size
        {
            get {
                long size = 0;
                foreach (DirectoryNode dir in m_SubDirs.Values)
                {
                    size += dir.Size;
                }
                return size + m_Files.SumFileSize;
            }
        }

        public Dictionary<string, DirectoryNode> Directories
        {
            get
            {
                return m_SubDirs;
            }
        }

        public string FileSystemPath
        {
            get
            {
                return m_FileSystemPath;
            }
            set
            {
                m_FileSystemPath = value;
            }
        }

        public string Name
        {
            get {
                return m_Name;
            }
            set
            {
                if (m_Name != value)
                {
                    if (Parent != null)
                    {
                        Remove();
                        Parent.Directories.Add(value.ToLower(), this);
                    }
                    m_Name = value;
                }
            }
        }

        public string FullVirtualPath
        {
            get
            {
                if (m_Parent == null)
                {
                    return string.Empty;
                }
                else
                {
                    return Path.Combine(m_Parent.FullVirtualPath, m_Name);
                }
            }
        }

        public FileDictionary Files
        {
            get
            {
                return m_Files;
            }
        }

        #endregion

        public DirectoryNode(FileLayoutManager manager, DirectoryNode parent, string name)
        {
            m_Parent = parent;
            m_Name = name;
            m_Manager = manager;
        }

        /// <summary>
        /// Returns the number of files and folders within the directory node.
        /// </summary>
        public int GetFileAndFolderCount()
        {
            int count = m_Files.Count;
            foreach (DirectoryNode dir in m_SubDirs.Values)
            {
                count++; // Count the current directory
                count += dir.GetFileAndFolderCount();
            }
            return count;
        }

        /// <summary>
        /// Removes this directory from the parent node.
        /// </summary>
        public bool Remove()
        {
            if (Parent != null)
            {
                Parent.Directories.Remove(Name.ToLower());
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Adds one or more paths from the file system to the directory.
        /// </summary>
        public ItemAddError[] AddPaths(string[] paths)
        {
            List<ItemAddError> errors = new List<ItemAddError>();

            foreach (string path in paths)
            {
                if (Manager.CancellationPending) break;

                if (string.IsNullOrEmpty(path)) continue;

                string normalisedPath = path.TrimEnd(Path.DirectorySeparatorChar);

                FileLayoutManager.CurrentProgressString = normalisedPath;

                try
                {
                    if (File.Exists(normalisedPath))
                    {
                        // Filter files that do not match the criteria
                        if (!Shell32.FileMatchesFilter(normalisedPath, ApplicationSettings.DataFilePattern, ApplicationSettings.DataFilePatternInclude))
                        {
                            continue;
                        }

                        FileNode node = new FileNode(normalisedPath, this, this.Manager.DataImage.VolumeType);
                        string key = node.FileInfo.Name.ToLower();
                        if (m_Files.ContainsKey(key))
                        {
                            FileLayoutManager.ReplaceAction action = Manager.RequestReplace(m_Files[key], node);
                            if (action == FileLayoutManager.ReplaceAction.Replace)
                            {
                                m_Files[key] = node;
                                if (!ReplaceItem(normalisedPath, node))
                                {
                                    throw new ApplicationException("Replacing the file failed for an unknown reason.");
                                }
                            }
                            else if (action == FileLayoutManager.ReplaceAction.Cancel)
                            {
                                return errors.ToArray();
                            }
                        }
                        else
                        {
                            if (!ReplaceItem(normalisedPath, node))
                            {
                                throw new ApplicationException("Replacing the file failed for an unknown reason.");
                            }
                            m_Files.Add(key, node);
                        }
                    }
                    else
                    {
                        // Assume directory
                        errors.AddRange(AddDirectory(normalisedPath));
                    }
                }
                catch (Exception ex)
                {
                    // Anything could happen...
                    errors.Add(new ItemAddError(normalisedPath, ex));
                }
            }

            return errors.ToArray();
        }

        private bool ReplaceItem(string sourceFile, FileNode node)
        {
            IFileLayoutX layout = Manager.DataImage.GetFileLayout();
            return layout.AddItem(sourceFile, node.FullVirtualPath);
        }

        /// <summary>
        /// Adds a directory structure to this node.
        /// </summary>
        /// <returns>Does not produce any errors, since no access to the 
        /// file system is involved.</returns>
        public ItemAddError[] AddDirectory(DirectoryNode dirAdd)
        {
            foreach (FileNode file in dirAdd.Files.Values)
            {
                string fileKey = file.Name.ToLower();
                if (Files.ContainsKey(fileKey))
                {
                    FileLayoutManager.ReplaceAction action = Manager.RequestReplace(Files[fileKey], file);
                    if (action == FileLayoutManager.ReplaceAction.Replace)
                    {
                        Manager.MoveDataImagePath(file, this);
                        Files[fileKey] = file;
                        file.Parent = this;
                    }
                    else if (action == FileLayoutManager.ReplaceAction.Cancel)
                    {
                        return new ItemAddError[0];
                    }
                }
                else
                {
                    Manager.MoveDataImagePath(file, this);
                    Files.Add(fileKey, file);
                    file.Parent = this;
                }
            }

            foreach (DirectoryNode dir in dirAdd.Directories.Values)
            {
                string dirKey = dir.Name.ToLower();
                dir.Parent = this;
                dir.m_Manager = this.m_Manager;
                if (Directories.ContainsKey(dirKey))
                {
                    Directories[dirKey].AddDirectory(dir);
                }
                else
                {
                   if (!Manager.DataImage.GetFileLayout().AddItem("", dir.FullVirtualPath))
                   {
                       throw new ApplicationException("Adding the directory failed for an unknown reason.");
                   }
                   Directories.Add(dirKey, dir);
                }
            }

            return new ItemAddError[0];
        }

        protected ItemAddError[] AddDirectory(string path)
        {
            List<ItemAddError> errors = new List<ItemAddError>();

            try
            {
                string dirName = Path.GetFileName(path);
                // Invalid source
                if (string.IsNullOrEmpty(dirName))
                {
                    return new ItemAddError[0]; 
                }

                string dirKey = dirName.ToLower();
                if (!Directories.ContainsKey(dirKey))
                {
                    DirectoryNode newDirNode = new DirectoryNode(Manager, this, dirName);
                    
                    IFileLayoutX layout = Manager.DataImage.GetFileLayout();
                    if (!layout.AddItem("", newDirNode.FullVirtualPath))
                    {
                        throw new ApplicationException("Adding the file failed for an unknown reason.");
                    }

                    Directories.Add(dirKey, newDirNode);
                    newDirNode.m_ImportedFileDate = Directory.GetLastWriteTime(path);
                    layout.GetLastItemAdded().SetFileTime(newDirNode.ModificationDate);
                }
                Directories[dirKey].FileSystemPath = path;

                List<string> dirObjects = new List<string>();
                dirObjects.AddRange(Directory.GetFileSystemEntries(path));
                errors.AddRange(Directories[dirKey].AddPaths(dirObjects.ToArray()));
            }
            catch (Exception ex)
            {
                errors.Add(new ItemAddError(path, ex));
            }

            return errors.ToArray();
        }

        /// <summary>
        /// Determines, whether or not the specified node is a subdirectory
        /// of this directory. If it is the same dir, it is also considered
        /// a subdirectory.
        /// </summary>
        public bool HasDescendant(DirectoryNode node)
        {
            if (this == node) return true;

            if (Directories.ContainsKey(node.Name.ToLower()))
            {
                return true;
            }
            else
            {
                foreach (DirectoryNode dirNode in Directories.Values)
                {
                    if (dirNode.HasDescendant(node))
                    {
                        return true;
                    }
                }
                return false;
            }
        }

        public override string ToString()
        {
            return FullVirtualPath;
        }
    }
}
