using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using CDBurnerXP.IO;

namespace CDBurnerXP.Controls
{
    public class ManagedDirectoryTreeView : TreeViewEx
    {
        private FileLayoutManager m_Manager = null;
        private bool m_SelectingDir = false;

        #region Properties

        public FileLayoutManager Manager
        {
            get
            {
                return m_Manager;
            }
            set
            {
                m_Manager = value;
            }
        }

        #endregion

        public ManagedDirectoryTreeView()
        {
            AllowDrop = true;
            LabelEdit = true;
            HideSelection = false;
            ImageList = FileLayoutManager.FileIcons;
            SelectedImageKey = "?FolderOpen";
            ShowRootLines = false;
            ImageKey = "?Folder";
        }

        protected override void OnDragDrop(DragEventArgs drgevent)
        {
            base.OnDragDrop(drgevent);

            TreeNode draggedTreeNode = null;
            DirectoryNode targetDirNode = SelectedNode.Tag as DirectoryNode;

            if (targetDirNode == null) return;

            m_Manager.StartReplace();

            if (drgevent.Data.GetDataPresent("ListViewItem"))
            {
                OLVListItem[] items = drgevent.Data.GetData("ListViewItem") as OLVListItem[];
                if (items == null) return;

                foreach (OLVListItem item in items)
                {
                    DirectoryNode dirNode = item.RowObject as DirectoryNode;
                    if (dirNode != null)
                    {
                        // If a DirNode has been dropped, simulate a drag and drop
                        // operation within the TreeView
                        draggedTreeNode = GetNode(dirNode.FullVirtualPath);
                        MoveDirectory(draggedTreeNode, SelectedNode);
                    }

                    FileNode fileNode = item.RowObject as FileNode;
                    if (fileNode != null && targetDirNode != null)
                    {
                        // Moving a file is rather simple, but show replace-dialog if necessary
                        string fileKey = fileNode.Name.ToLower();
                        if (targetDirNode.Files.ContainsKey(fileKey))
                        {
                            FileLayoutManager.ReplaceAction action = m_Manager.RequestReplace(targetDirNode.Files[fileKey], fileNode);
                            if (action == FileLayoutManager.ReplaceAction.Replace)
                            {
                                m_Manager.MoveDataImagePath(fileNode, targetDirNode);
                                fileNode.Remove();
                                targetDirNode.Files[fileKey] = fileNode;
                                fileNode.Parent = targetDirNode;
                            }
                            else if (action == FileLayoutManager.ReplaceAction.Cancel)
                            {
                                m_Manager.SetCompilationChanged(false);
                                return;
                            }
                        }
                        else
                        {
                            m_Manager.MoveDataImagePath(fileNode, targetDirNode);
                            fileNode.Remove();
                            targetDirNode.Files.Add(fileKey, fileNode);
                            fileNode.Parent = targetDirNode;
                        }
                    }
                }

                ShowDirContents(targetDirNode);
                m_Manager.SetCompilationChanged(false);

                return;
            }
            else if (drgevent.Data.GetDataPresent("TreeNode"))
            {
                draggedTreeNode = drgevent.Data.GetData("TreeNode") as TreeNode;
            }
            else if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] paths = drgevent.Data.GetData(DataFormats.FileDrop) as string[];
                m_Manager.BeginAddPaths(targetDirNode.FullVirtualPath, paths);
                RefreshDirContents();
            }

            MoveDirectory(draggedTreeNode, SelectedNode);
        }

        /// <summary>
        /// Moves one directory to other other, determined by the
        /// corresponding TreeNodes.
        /// </summary>
        public void MoveDirectory(TreeNode draggedTreeNode, TreeNode targetTreeNode)
        {
            if (targetTreeNode == null) return;

            DirectoryNode targetDirNode = targetTreeNode.Tag as DirectoryNode;

            if (draggedTreeNode == null || targetDirNode == null) return;

            // Note: Can't move parent node into its childs
            if (IsDescendant(draggedTreeNode, targetTreeNode)) return;

            // If dragged to the same location, nothing to do
            if (draggedTreeNode.Parent == targetTreeNode) return;

            // Reassign directory nodes in the virtual structure
            DirectoryNode draggedDirNode = draggedTreeNode.Tag as DirectoryNode;
            string dirKey = draggedDirNode.Name.ToLower();
            draggedDirNode.Remove();

            // Directory could already exist at the specified location!
            if (targetDirNode.Directories.ContainsKey(dirKey))
            {
                DirectoryNode existing = targetDirNode.Directories[dirKey];
                existing.AddDirectory(draggedDirNode);
                draggedTreeNode.Remove();
                UpdateNode(targetTreeNode);
            }
            else
            {
                m_Manager.MoveDataImagePath(draggedDirNode, targetDirNode);
                targetDirNode.Directories.Add(dirKey, draggedDirNode);
                draggedDirNode.Parent = targetDirNode;
                // Reassign nodes in TreeView
                draggedTreeNode.Remove();
                targetTreeNode.Nodes.Add(draggedTreeNode);
            }

            m_Manager.SetCompilationChanged(false);

            // Keep selection, sort, expand
            TreeNode oldSelection = SelectedNode;
            targetTreeNode.Expand();
            Sort();
            SelectedNode = oldSelection;
        }

        /// <summary>
        /// Goes through the specified node and subnodes, compares
        /// them with the directory structure, and adds all missing
        /// nodes.
        /// </summary>
        public void UpdateNode()
        {
            UpdateNode(string.Empty);
        }

        /// <summary>
        /// Goes through the specified node and subnodes, compares
        /// them with the directory structure, and adds all missing
        /// nodes.
        /// </summary>
        public void UpdateNode(string path)
        {
            UpdateNode(GetNode(path));
        }

        /// <summary>
        /// Goes through the specified node and subnodes, compares
        /// them with the directory structure, and adds all missing
        /// nodes.
        /// </summary>
        public void UpdateNode(TreeNode node)
        {
            if (InvokeRequired)
            {
                this.Invoke(new MethodInvoker(delegate() { UpdateNode(node); }));
                return;
            }

            TreeNode oldSelection = SelectedNode;

            BeginUpdate();

            UpdateNodeRecurse(node);

            Sort();
            
            // Don't forget to restore selection
            EndUpdate();
            if (oldSelection != null)
            {
                SelectedNode = oldSelection;
            }
        }

        private void UpdateNodeRecurse(TreeNode node)
        {
            Application.DoEvents();
            DirectoryNode dirNode = node.Tag as DirectoryNode;
            foreach (DirectoryNode dir in dirNode.Directories.Values)
            {
                string dirKey = dir.Name.ToLower();
                if (!node.Nodes.ContainsKey(dirKey))
                {
                    TreeNode newNode = node.Nodes.Add(dirKey, dir.Name);
                    if (dir.IsImported)
                    {
                        newNode.ImageKey = "?FolderImport";
                        newNode.SelectedImageKey = "?FolderOpenImport";
                    }
                    newNode.Tag = dir;
                    UpdateNodeRecurse(newNode);
                }
                else
                {
                    UpdateNodeRecurse(node.Nodes[dirKey]);
                }
            }
        }

        protected bool IsDescendant(OLVListItem root, TreeNode child)
        {
            if (root.RowObject is FileNode) return false; // Files can be dropped anywhere

            DirectoryNode dirNode = root.RowObject as DirectoryNode;

            TreeNode node = GetNode(dirNode.FullVirtualPath);
            return IsDescendant(node, child);
        }

        protected bool IsDescendant(TreeNode root, TreeNode child)
        {
            DirectoryNode rootDir = root.Tag as DirectoryNode;
            DirectoryNode childDir = child.Tag as DirectoryNode;
            return rootDir.HasDescendant(childDir);
        }

        protected override void OnItemDrag(ItemDragEventArgs e)
        {
            base.OnItemDrag(e);

            TreeNode node = e.Item as TreeNode;
            if (node == null) return;
            
            if (e.Button == MouseButtons.Left)
            {
                DirectoryNode dirNode = node.Tag as DirectoryNode;
                if (dirNode == null || dirNode.IsImported) return;

                DoDragDrop(new DataObject("TreeNode", e.Item), DragDropEffects.Copy | DragDropEffects.Move);
            }
        }

        protected override void OnDragOver(DragEventArgs drgevent)
        {
            base.OnDragOver(drgevent);
            
            Point location = PointToClient(new Point(drgevent.X, drgevent.Y));
            TreeNode node = GetNodeAt(location);
            if (node != null && node.Tag != null)
            {
                TreeNode draggedNode = drgevent.Data.GetData("TreeNode") as TreeNode;
                // Indicate, that dropping to the descending nodes is not allowed
                if (draggedNode != null)
                {
                    if (IsDescendant(draggedNode, node))
                    {
                        drgevent.Effect = DragDropEffects.None;
                    }
                    else
                    {
                        // If OK, reset the cursor
                        OnDragEnter(drgevent);
                    }
                }

                // Again, check if drag and drop is OK
                OLVListItem[] draggedItems = drgevent.Data.GetData("ListViewItem") as OLVListItem[];
                if (draggedItems != null)
                {
                    foreach (OLVListItem draggedItem in draggedItems)
                    {
                        if (IsDescendant(draggedItem, node))
                        {
                            drgevent.Effect = DragDropEffects.None;
                            break;
                        }
                        else
                        {
                            // If OK, reset the cursor
                            OnDragEnter(drgevent);
                        }
                    }
                }

                SelectedNode = node;
            }
            else
            {
                SelectedNode = Nodes[0];
                OnDragEnter(drgevent);
            }

            // While doing stuff in the TreeView, we should not indicate
            // anything in the ListView (might appear like a drop target)
            m_Manager.ListView.SelectedIndices.Clear();
        }

        protected override void OnDragEnter(DragEventArgs drgevent)
        {
            base.OnDragEnter(drgevent);

            if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
            {
                drgevent.Effect = DragDropEffects.Copy;
            }
            else if (drgevent.Data.GetDataPresent("TreeNode"))
            {
                drgevent.Effect = DragDropEffects.Move;
            }
            else if (drgevent.Data.GetDataPresent("ListViewItem"))
            {
                drgevent.Effect = DragDropEffects.Move;
            }
        }

        protected override void OnAfterSelect(TreeViewEventArgs e)
        {
            base.OnAfterSelect(e);
            
            if (m_SelectingDir) return;

            OpenDirectory(GetSelectedPath(), false);
        }

        protected override void OnBeforeCollapse(TreeViewCancelEventArgs e)
        {
            base.OnBeforeCollapse(e);

            // Root node cannot be collapsed
            if (e.Node == Nodes[0]) e.Cancel = true;
        }

        /// <summary>
        /// Shows the directory's contents of the currently
        /// selected node in the ListView.
        /// </summary>
        public void RefreshDirContents()
        {
            if (SelectedNode == null)
            {
                ShowDirContents(Nodes[0].Tag as DirectoryNode);
            }
            else
            {
                DirectoryNode dirNode = SelectedNode.Tag as DirectoryNode;
                ShowDirContents(dirNode);
            }
        }

        /// <summary>
        /// Opens the parent directory of the current directory.
        /// </summary>
        public void OpenParentDirectory()
        {
            TreeNode node = SelectedNode;
            if (node != null && node.Parent != null)
            {
                SelectedNode = node.Parent;
            }
        }

        /// <summary>
        /// Opens the selected path. The corresponding node
        /// in the TreeView will be selected and its contents shown.
        /// </summary>
        /// <remarks>If the directory does not exist, nothing happens.</remarks>
        /// <param name="path">Path to show</param>
        /// <param name="expand">Whether or not to expand the node after selection.</param>
        public void OpenDirectory(string path, bool expand)
        {
            TreeNode node = GetNode(path);
            if (node == null) return;

            DirectoryNode dirNode = node.Tag as DirectoryNode;
            if (dirNode == null) return;

            m_SelectingDir = true;
            SelectedNode = node;
            m_SelectingDir = false;
            
            ShowDirContents(dirNode);
            if (expand) SelectedNode.Expand();
        }

        /// <summary>
        /// Shows the directory's contents of the specified node
        /// in the ListView.
        /// </summary>
        private void ShowDirContents(DirectoryNode node)
        {
            // Every time the TreeView is "used", vheck whether or not root node
            // is expanded (should always be the case
            if (Nodes[0] != null && !Nodes[0].IsExpanded) Nodes[0].Expand();

            List<IFileSystemNode> objects = new List<IFileSystemNode>();
            foreach (DirectoryNode dirNode in node.Directories.Values)
            {
                objects.Add(dirNode);
            }
            foreach (FileNode fileNode in node.Files.Values)
            {
                objects.Add(fileNode);
            }

            m_Manager.ListView.SetObjects(objects);
            m_Manager.ListView.Refresh();
        }

        protected override void OnAfterLabelEdit(NodeLabelEditEventArgs e)
        {
            base.OnAfterLabelEdit(e);

            if (e.Node == null) return;

            DirectoryNode dirNode = e.Node.Tag as DirectoryNode;

            if (e.Label == null)
            {
                // Did we create a new folder?
                if (dirNode.VirtualDirectory)
                {
                    // Detele folder
                    e.Node.Remove();
                }
                // Rename or creation cancelled
                return;
            }
            
            // We do our own stuff here
            e.CancelEdit = true;

            // Did we create a new folder?
            if (dirNode.VirtualDirectory)
            {
                if (m_Manager.Rename(dirNode.FullVirtualPath, e.Label, dirNode))
                {
                    string pathToNewNode = System.IO.Path.Combine(dirNode.Parent.FullVirtualPath, e.Label);
                    // All OK, delete virtual folder and create a new one
                    BeginUpdate();
                    m_Manager.CreateDirectory(pathToNewNode);
                    e.Node.Remove();
                    SelectedNode = GetNode(pathToNewNode);
                    EndUpdate();
                }
                else
                {
                    // Could not add directory
                    e.Node.BeginEdit();
                }
            }
            else
            {
                m_Manager.Rename(dirNode.FullVirtualPath, e.Label);
            }
        }

        private delegate TreeNode GetNodeDelegate();

        /// <summary>
        /// Determines the path of a given TreeNode.
        /// </summary>
        public string GetPath(TreeNode node)
        {
            if (node == null) return string.Empty;
            
            DirectoryNode dirNode = node.Tag as DirectoryNode;
            if (dirNode == null) return string.Empty;

            return dirNode.FullVirtualPath;
        }

        /// <summary>
        /// Determines the TreeView node for the specified
        /// path.
        /// </summary>
        /// <returns>null if node not found</returns>
        public TreeNode GetNode(string path)
        {
            if (InvokeRequired)
            {
                return this.Invoke(new GetNodeDelegate(delegate() { return GetNode(path); })) as TreeNode;
            }

            if (string.IsNullOrEmpty(path)) return Nodes[0]; // root

            string[] parts = FileLayoutManager.SplitPath(path);

            TreeNode current = Nodes[0];
            foreach (string part in parts)
            {
                if (!current.Nodes.ContainsKey(part))
                {
                    return null;
                }
                current = current.Nodes[part];
            }
            return current;
        }

        /// <summary>
        /// Determines the path of the currently selected
        /// (shown in listview) directory.
        /// </summary>
        /// <returns></returns>
        public string GetSelectedPath()
        {
            TreeNode node = (SelectedNode == null) ? Nodes[0] : SelectedNode;
            DirectoryNode dirNode = node.Tag as DirectoryNode;
            return dirNode.FullVirtualPath;
        }

        public void DeleteSelection()
        {
            if (SelectedNode != null && m_Manager != null)
            {
                DirectoryNode dirNode = SelectedNode.Tag as DirectoryNode;
                m_Manager.Delete(new string[] { dirNode.FullVirtualPath });
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Delete:
                    DeleteSelection();
                    break;
                case Keys.F2:
                    if (SelectedNode != null)
                    {
                        SelectedNode.BeginEdit();
                    }
                    break;
            }

            base.OnKeyDown(e);
        }

    }
}
