using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using CDBurnerXP.IO;
using CDBurnerXP.Printing;
using CDBurnerXP.Text;
using System.IO;
using System.Drawing;
using Microsoft.Win32;
using System.Collections;
using System.ComponentModel;
using CDBurnerXP.Forms;
using CDBurnerXP.Drawing;

namespace CDBurnerXP.Controls
{
    public class ManagedFileListView : FastObjectListView
    {
        private FileLayoutManager m_Manager = null;
        private bool m_CancelNextCellEdit = false;
        private Control m_CellEditor = null;

        #region LogicalFileSorter

        /// <summary>
        /// Sorts files in the list view like Windows (XP and later) does, considering numeric file parts.
        /// </summary>
        private class LogicalFileSorter : ModelObjectComparer
        {
            public LogicalFileSorter(OLVColumn col, SortOrder order, OLVColumn col2, SortOrder order2)
                : base(col, order, col2, order2)
            {
            }

            public override int Compare(object x, object y)
            {
                IFileSystemNode node1 = x as IFileSystemNode;
                IFileSystemNode node2 = y as IFileSystemNode;

                // Windows behaviour: Group directories at the top or bottom
                if (node1 != null && node2 != null)
                {
                    if (x is DirectoryNode && !(y is DirectoryNode))
                    {
                        return this.sortOrder == SortOrder.Ascending ? -1 : 1;
                    }

                    if (y is DirectoryNode && !(x is DirectoryNode))
                    {
                        return this.sortOrder == SortOrder.Ascending ? 1 : -1;
                    }
                }

                if (node1 != null && node2 != null && this.column.Name == "cName")
                {
                    return (this.sortOrder == SortOrder.Ascending) ? StringLogicalComparer.Compare(node1.Name, node2.Name) : StringLogicalComparer.Compare(node2.Name, node1.Name);
                }

                return base.Compare(x, y);
            }
        }

        #endregion

        #region Properties

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public new string EmptyListMsg
        {
            get
            {
                return base.EmptyListMsg;
            }
            set
            {
                base.EmptyListMsg = value;
            }
        }

        public FileLayoutManager Manager
        {
            get
            {
                return m_Manager;
            }
            set
            {
                m_Manager = value;
            }
        }

        #endregion

        #region Delegate functions

        protected object ShowFileIcon(object x)
        {
            ImageList.ImageCollection images = FileLayoutManager.FileIcons.Images;

            DirectoryNode dirNode = x as DirectoryNode;
            if (dirNode != null)
            {
                return (dirNode.IsImported ? images.IndexOfKey("?FolderImport") : images.IndexOfKey("?Folder"));
            }

            FileNode fileNode = x as FileNode;
            if (fileNode != null)
            {
                // Need to determine file icon?
                string ext = Path.GetExtension(fileNode.Name);
                if (!fileNode.IsImported && !string.IsNullOrEmpty(ext))
                {
                    try
                    {
                        string key = ext.ToLower();
                        // Exe and link files do not all have the same extension
                        if (key == ".lnk" || key == ".exe")
                        {
                            key = fileNode.FileSystemPath;
                        }
                        if (!images.ContainsKey(key))
                        {
                            Icon myIcon = IconReader.GetFileIcon(fileNode.FileSystemPath, IconReader.IconSize.Small, false);
                            images.Add(key, RtlHelper.FlipIfNecessary(myIcon, BaseForm.IsRightToLeftCulture));
                        }
                        return images.IndexOfKey(key);
                    }
                    catch { /* return defaults */ }
                }
                return (fileNode.IsImported) ? images.IndexOfKey("?FileImport") : images.IndexOfKey("?File");
            }

            return 0;
        }

        protected object GetFileSize(object x)
        {
            IFileSystemNode node = x as IFileSystemNode;
            if (node != null)
            {
                return node.Size;
            }
            return 0L;
        }

        protected string ShowFileSize(object x)
        {
            long size = Convert.ToInt64(x);

            return FormatFileSize.Format(size);
        }

        protected object GetFileDate(object x)
        {
            IFileSystemNode node = x as IFileSystemNode;
            if (node != null)
            {
                try
                {
                    return node.ModificationDate;
                }
                catch (ArgumentOutOfRangeException)
                {
                    // Not a valid Win32 FileTime (can occur sometimes)
                }
            }
            return DateTime.MinValue;
        }

        protected string ShowFileType(object x)
        {
            IFileSystemNode node = x as IFileSystemNode;
            if (node == null) return string.Empty;

            if (node is DirectoryNode) return FileLayoutManager.Translation[FileLayoutManager.UIText.Folder];

            string ext = Path.GetExtension(node.Name);

            if (string.IsNullOrEmpty(ext))
            {
                return FileLayoutManager.Translation[FileLayoutManager.UIText.File];
            }
            else
            {
                string fileType = PathEx.GetFileTypeFromExt(ext);
                if (string.IsNullOrEmpty(fileType))
                {
                    return ext.Trim('.') + " " + FileLayoutManager.Translation[FileLayoutManager.UIText.File];
                }
                return fileType;                
            }
        }

        protected string ShowFileDate(object x)
        {
            DateTime value = (DateTime)x;
            if (value != DateTime.MinValue)
            {
                return value.ToString("g");
            }
            return string.Empty;
        }

        protected object ShowFileItem(object x)
        {
            FileNode node = x as FileNode;
            if (node != null)
            {
                return Path.GetFileName(node.FullVirtualPath);
            }
            DirectoryNode dirNode = x as DirectoryNode;
            if (dirNode != null)
            {
                return dirNode.Name;
            }
            return string.Empty;
        }

        #endregion

        public ManagedFileListView() : base()
        {
            CellEditActivation = CellEditActivateMode.DoubleClick;
            AllowDrop = true;
            HideSelection = false;
            AllowColumnReorder = true;

            AllColumns.Clear();
            SmallImageList = FileLayoutManager.FileIcons;

            // Name
            OLVColumn col = new OLVColumn();
            col.Name = "cName";
            col.Text = FileLayoutManager.Translation[FileLayoutManager.UIText.Name];
            col.AspectGetter = ShowFileItem;
            col.ImageGetter = ShowFileIcon;
            col.Width = 150;
            AllColumns.Add(col);

            // Size
            col = new OLVColumn();
            col.Name = "cSize";
            col.Text = FileLayoutManager.Translation[FileLayoutManager.UIText.Size];
            col.AspectGetter = GetFileSize;
            col.AspectToStringConverter = ShowFileSize;
            col.Width = 100;
            col.IsEditable = false;
            col.TextAlign = HorizontalAlignment.Right;
            AllColumns.Add(col);

            // Type
            col = new OLVColumn();
            col.Name = "cType";
            col.Text = FileLayoutManager.Translation[FileLayoutManager.UIText.Type];
            col.AspectGetter = ShowFileType;
            col.IsEditable = false;
            col.Width = 100;
            AllColumns.Add(col);

            // Date
            col = new OLVColumn();
            col.Name = "cDate";
            col.Text = FileLayoutManager.Translation[FileLayoutManager.UIText.Date];
            col.AspectGetter = GetFileDate;
            col.IsEditable = false;
            col.Width = 100;
            col.AspectToStringConverter = ShowFileDate;
            AllColumns.Add(col);

            RebuildColumns();

            EmptyListMsg = FileLayoutManager.Translation[FileLayoutManager.UIText.AddFilesHere];

            this.CustomSorter = delegate(OLVColumn column, SortOrder sortOrder)
            {
                this.ClearCachedInfo();
                if (sortOrder != SortOrder.None)
                    this.Objects.Sort(new LogicalFileSorter(column, sortOrder, this.SecondarySortColumn, this.SecondarySortOrder));
                this.RebuildIndexMap();
            };
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            // Make sure the cell editor gets it and that it's not consumed by menu shortcuts
            if (IsCellEditing && m_CellEditor != null)
            {
                TextBox textBox = m_CellEditor as TextBox;

                switch (keyData)
                {
                    case Keys.Delete:
                        User32.SendMessage(m_CellEditor.Handle, (uint)msg.Msg, msg.WParam, msg.LParam);
                        return true;

                    case Keys.Control | Keys.V:
                        if (textBox != null) textBox.Paste();
                        return true;

                    case Keys.Control | Keys.C:
                        if (textBox != null) textBox.Copy();
                        return true;
                        
                    case Keys.F2:
                        // Do nothing. Having this done twice is not good.
                        return true;
                }
            }

            if (!IsCellEditing)
            {
                switch (keyData)
                {
                    case Keys.Delete:
                        DeleteSelection();
                        return true;

                    case (Keys.Control | Keys.V):
                        if (Paste())
                        {
                            return true;
                        }
                        break;

                    case Keys.Enter:
                        DirectoryNode item = SelectedObject as DirectoryNode;
                        if (item == null) return false;

                        m_Manager.TreeView.OpenDirectory(item.FullVirtualPath, true);
                        break;

                    case Keys.Back:
                        m_Manager.TreeView.OpenParentDirectory();
                        break;

                    case Keys.Alt | Keys.Enter:
                        Manager.ShowProperties();
                        return true;
                }
            }

            return base.ProcessCmdKey(ref msg, keyData);
        }

        protected override void OnCellEditStarting(ObjectListView.CellEditEventArgs e)
        {
            m_CellEditor = e.Control;

            base.OnCellEditStarting(e);
            
            // Prevent renaming when double clicking on directories
            if (m_CancelNextCellEdit)
            {
                e.Cancel = true;
                m_CancelNextCellEdit = false;
            }

            TextBox txt = e.Control as TextBox;
            if (txt != null)
            {
                txt.AutoCompleteMode = AutoCompleteMode.None;
            }
        }


        protected override void OnCellEditorValidating(ObjectListView.CellEditEventArgs e)
        {
            base.OnCellEditorValidating(e);

            DirectoryNode dirNode = e.RowObject as DirectoryNode;

            if (dirNode != null && dirNode.VirtualDirectory)
            {
                // Verify that directory can be created
                if (!m_Manager.Rename(dirNode.FullVirtualPath, e.Control.Text, dirNode))
                {
                    e.Cancel = true;
                }
            }
            else
            {
                IFileSystemNode node = e.RowObject as IFileSystemNode;
                e.Cancel = !m_Manager.Rename(node.FullVirtualPath, e.Control.Text);
            }
        }

        protected override void OnCellEditFinishing(ObjectListView.CellEditEventArgs e)
        {
            base.OnCellEditFinishing(e);
            
            DirectoryNode dirNode = e.RowObject as DirectoryNode;

            if (dirNode != null && dirNode.VirtualDirectory)
            {
                // Create the new virtual directory
                string pathToNewNode = Path.Combine(dirNode.Parent.FullVirtualPath, e.Control.Text);
                m_Manager.CreateDirectory(pathToNewNode);
                m_Manager.TreeView.RefreshDirContents();
                SelectObject(m_Manager.GetNode(pathToNewNode));
            }

            m_CellEditor = null;
        }
        
        protected override void OnItemDrag(ItemDragEventArgs e)
        {
            base.OnItemDrag(e);

            List<OLVListItem> items = new List<OLVListItem>();
            foreach (int index in SelectedIndices)
            {
                items.Add(GetItem(index));
            }

            if (items.Count == 0) return;

            // Check that no imported items are included
            foreach (OLVListItem item in items)
            {
                if (((IFileSystemNode)item.RowObject).IsImported) return;
            }

            if (e.Button == MouseButtons.Left)
            {
                DoDragDrop(new DataObject("ListViewItem", items.ToArray()), DragDropEffects.Copy | DragDropEffects.Move);
            }
        }

        public override void SetObjects(IEnumerable collection)
        {
            if (collection == null)
            {
                CleanupCellEdit();
            }
            base.SetObjects(collection);
        }

        protected override void OnDragOver(DragEventArgs drgevent)
        {
            base.OnDragOver(drgevent);

            Point location = PointToClient(new Point(drgevent.X, drgevent.Y));
            OLVListItem targetItem = GetItemAt(location.X, location.Y) as OLVListItem;
            DirectoryNode targetDirNode = null;

            if (targetItem == null) {
                // If no target folder has been selected, use the currently shown directory
                if (m_Manager.TreeView.SelectedNode == null)
                {
                    targetDirNode = m_Manager.TreeView.Nodes[0].Tag as DirectoryNode;
                }
                else
                {
                    targetDirNode = m_Manager.TreeView.SelectedNode.Tag as DirectoryNode;
                }
            } else {
                targetDirNode = targetItem.RowObject as DirectoryNode;
            }

            // We can only drop on folders within the listview
            if (targetDirNode == null)
            {
                drgevent.Effect = DragDropEffects.None;
                return;
            }

            OLVListItem[] draggingItems = drgevent.Data.GetData("ListViewItem") as OLVListItem[];
            if (draggingItems != null)
            {
                bool ok = true;

                foreach (OLVListItem draggingItem in draggingItems)
                {
                    // Dragging files is always OK
                    DirectoryNode draggingDir = draggingItem.RowObject as DirectoryNode;
                    if (draggingDir != null && (draggingDir.HasDescendant(targetDirNode) || draggingDir == targetDirNode))
                    {
                        ok = false;
                    }
                }

                if (ok)
                {
                    SelectedIndices.Clear();
                    SelectedObject = targetDirNode;
                    drgevent.Effect = DragDropEffects.Move;
                    return;
                }
            }

            TreeNode draggingTreeItem = drgevent.Data.GetData("TreeNode") as TreeNode;
            if (draggingTreeItem != null)
            {
                DirectoryNode draggingDir = draggingTreeItem.Tag as DirectoryNode;
                if (!draggingDir.HasDescendant(targetDirNode) && draggingDir != targetDirNode)
                {
                    SelectedIndices.Clear();
                    SelectedObject = targetDirNode;
                    drgevent.Effect = DragDropEffects.Move;
                    return;
                }
            }

            string[] draggedPaths = drgevent.Data.GetData(DataFormats.FileDrop) as string[];
            if (draggedPaths != null)
            {
                SelectedIndices.Clear();
                SelectedObject = targetDirNode;
                drgevent.Effect = DragDropEffects.Copy;
                return;
            }

            drgevent.Effect = DragDropEffects.None;
        }

        protected override void OnDragDrop(DragEventArgs drgevent)
        {
            base.OnDragDrop(drgevent);

            OLVListItem[] draggedItems = drgevent.Data.GetData("ListViewItem") as OLVListItem[];
            DirectoryNode targetDirNode = SelectedObject as DirectoryNode;

            if (targetDirNode == null)
            {
                // If no target folder has been selected, use the currently shown directory
                if (m_Manager.TreeView.SelectedNode == null)
                {
                    targetDirNode = m_Manager.TreeView.Nodes[0].Tag as DirectoryNode;
                }
                else
                {
                    targetDirNode = m_Manager.TreeView.SelectedNode.Tag as DirectoryNode;
                }
            }

            if (draggedItems != null)
            {
                m_Manager.StartReplace();

                foreach (OLVListItem draggedItem in draggedItems)
                {
                    // We'll either have a file or directory here
                    FileNode draggedFile = draggedItem.RowObject as FileNode;
                    if (draggedFile != null && draggedFile.Parent != targetDirNode)
                    {
                        string fileKey = draggedFile.Name.ToLower();
                        // Check if file actually exists at the specified location
                        if (targetDirNode.Files.ContainsKey(fileKey))
                        {
                            FileLayoutManager.ReplaceAction action = m_Manager.RequestReplace(draggedFile, targetDirNode.Files[fileKey]);
                            if (action == FileLayoutManager.ReplaceAction.Replace)
                            {
                                m_Manager.MoveDataImagePath(draggedFile, targetDirNode);
                                draggedFile.Remove();
                                targetDirNode.Files[fileKey] = draggedFile;
                            }
                            else if (action == FileLayoutManager.ReplaceAction.Keep)
                            {
                                continue;
                            }
                            else if (action == FileLayoutManager.ReplaceAction.Cancel)
                            {
                                m_Manager.SetCompilationChanged(false);
                                return;
                            }
                        }
                        else
                        {
                            m_Manager.MoveDataImagePath(draggedFile, targetDirNode);
                            draggedFile.Remove();
                            targetDirNode.Files.Add(fileKey, draggedFile);
                        }

                        m_Manager.SetCompilationChanged(false);
                        draggedFile.Parent = targetDirNode;
                        m_Manager.TreeView.RefreshDirContents();
                    }

                    DirectoryNode draggedDir = draggedItem.RowObject as DirectoryNode;
                    if (draggedDir != null)
                    {
                        // A dir has been dragged. Nice, the TreeView can handle that for us.
                        TreeNode targetNode = m_Manager.TreeView.GetNode(targetDirNode.FullVirtualPath);
                        m_Manager.TreeView.MoveDirectory(m_Manager.TreeView.GetNode(draggedDir.FullVirtualPath), targetNode);
                    }
                }
            }

            TreeNode draggedNode = drgevent.Data.GetData("TreeNode") as TreeNode;
            if (draggedNode != null)
            {
                // Let the TreeView handle drops of TreeNodes
                TreeNode targetNode = m_Manager.TreeView.GetNode(targetDirNode.FullVirtualPath);
                m_Manager.TreeView.MoveDirectory(draggedNode, targetNode);
            }

            string[] paths = drgevent.Data.GetData(DataFormats.FileDrop) as string[];
            if (paths != null)
            {
                m_Manager.BeginAddPaths(targetDirNode.FullVirtualPath, paths);
                m_Manager.TreeView.RefreshDirContents();
            }
        }

        protected override void OnDragEnter(DragEventArgs drgevent)
        {
            base.OnDragEnter(drgevent);

            if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
            {
                drgevent.Effect = DragDropEffects.Copy;
            }
            else if (drgevent.Data.GetDataPresent("ListViewItem"))
            {
                drgevent.Effect = DragDropEffects.Move;
            }
        }

        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            base.OnMouseDoubleClick(e);

            OLVListItem item = GetItemAt(e.X, e.Y) as OLVListItem;
            // Open directory on double click
            if (item != null && e.Clicks == 2 && e.Button == MouseButtons.Left)
            {
                DirectoryNode dirNode = item.RowObject as DirectoryNode;
                if (dirNode != null && m_Manager.TreeView != null)
                {
                    m_CancelNextCellEdit = true;
                    m_Manager.TreeView.OpenDirectory(dirNode.FullVirtualPath, true);
                }
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);
            // The OLV has - for technical reasons - a different
            // understanding of double clicks than we have. So to 
            // prevent conflicts with m_CancelNextCellEdit, we make
            // sure it is properly reset at this occasion as well.
            m_CancelNextCellEdit = false;
        }

        public CoverListItem[] GetCoverFileList()
        {
            if (Objects == null) return new CoverListItem[0];

            List<CoverListItem> items = new List<CoverListItem>();

            foreach (IFileSystemNode node in Objects)
            {
                string path = node.FileSystemPath ?? node.Name;
                items.Add(new CoverListItem(node.Name, path, FormatFileSize.Format(node.FileSystemPath)));
            }

            return items.ToArray();
        }

        #region UI functions

        public void DeleteSelection()
        {
            if (IsCellEditing) return;

            List<string> paths = new List<string>();
            foreach (IFileSystemNode node in SelectedObjects)
            {
                paths.Add(node.FullVirtualPath);
            }

            m_Manager.Delete(paths.ToArray());
            DeselectAll();
        }

        public bool Paste()
        {
            if (IsCellEditing) return false;
            string[] paths = SafeClipboard.GetData(DataFormats.FileDrop) as string[];
            if (paths == null) return false;

            TreeNode currentDir = (m_Manager.TreeView.SelectedNode == null) ? m_Manager.TreeView.Nodes[0] : m_Manager.TreeView.SelectedNode;

            // Check where we paste to
            DirectoryNode dirNode = currentDir.Tag as DirectoryNode;

            m_Manager.BeginAddPaths(dirNode.FullVirtualPath, paths);
            return true;
        }

        public void InvertSelection()
        {
            int[] selectedIndices = new int[SelectedIndices.Count];
            SelectedIndices.CopyTo(selectedIndices, 0);
            List<int> listSelectedIndices = new List<int>(selectedIndices);

            DeselectAll();

            for (int i = 0; i < Items.Count; i++) {
                if (!listSelectedIndices.Contains(i))
                {
                    SelectedIndices.Add(i);
                }
            }
        }
        
        #endregion
    }
}
