using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using CDBurnerXP.IO;
using RocketDivision.StarBurnX;
using System.Runtime.InteropServices;

namespace CDBurnerXP.Forms
{
    public partial class AddingAudioTracks : BaseForm
    {
        #region ProgressAddingTrack

        protected enum Status
        {
            None = -1,
            Working = 0,
            Success = 1,
            Failure = 2
        }

        protected class ProgressAddingTrack
        {
            private Status m_Status = Status.None;
            private CDBurnerXP.Controls.AudioTrack m_AudioTrack = null;
            private short m_Progress = 0;

            public CDBurnerXP.Controls.AudioTrack AudioTrack
            {
                get { return m_AudioTrack; }
            }

            public Status Status
            {
                get
                {
                    return m_Status;
                }
                set
                {
                    m_Status = value;
                }
            }

            public string FileName
            {
                get
                {
                    return m_AudioTrack.FileName;
                }
            }

            public short Progress
            {
                get
                {
                    return m_Progress;
                }
                set
                {
                    m_Progress = value;
                }
            }

            public ProgressAddingTrack(CDBurnerXP.Controls.AudioTrack audioTrack)
            {
                m_AudioTrack = audioTrack;
            }
        }

        #endregion

        protected CDBurnerXP.Controls.AudioTrack[] m_Tracks = null;
        private AudioBurner m_AudioImage = null;
        protected List<ProgressAddingTrack> m_WorkItems = new List<ProgressAddingTrack>();
        protected ProgressAddingTrack m_CurrentTrack = null;
        private CDBurnerXP.Controls.AudioLayoutManager m_Layout = null;

        #region Properties

        protected override bool AutoTranslate
        {
            get
            {
                return true;
            }
        }

        #endregion

        protected AddingAudioTracks() {
            InitializeComponent();
            CancelButton = bCancel;

            cFile.AspectGetter = delegate(object x) { return Path.GetFileName(((ProgressAddingTrack)x).FileName); };
            cFile.ImageGetter = delegate(object x) { return Convert.ToInt32(((ProgressAddingTrack)x).Status); };
            cProgress.AspectGetter = delegate(object x) { return ((ProgressAddingTrack)x).Progress; };
            cProgress.Renderer = new CDBurnerXP.Controls.BarRenderer(0, 100);
        }

        public AddingAudioTracks(CDBurnerXP.Controls.AudioLayoutManager layout, AudioBurner audioImage, string targetPath) : this()
        {
            m_Layout = layout;
            m_Tracks = layout.AudioTracks;
            m_AudioImage = audioImage;
            AudioConverter.TempPath = targetPath;
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            if (DesignMode) return;

            cFile.Text = RM.GetString(cFile.Text);
            cProgress.Text = RM.GetString(cProgress.Text);

            foreach (CDBurnerXP.Controls.AudioTrack track in m_Tracks)
            {
                m_WorkItems.Add(new ProgressAddingTrack(track));
            }

            olvProgress.SetObjects(m_WorkItems);
            // Prepare and start worker
            AudioConverter.ProgressChanged += new EventHandler<ProgressChangedEventArgs>(AudioConverter_ProgressChanged);
            if (m_AudioImage != null)
            {
                m_AudioImage.OnProgress += new IAudioBurnerEvents_OnProgressEventHandler(m_AudioImage_OnProgress);
            }
            m_Worker.RunWorkerAsync();
        }

        /// <summary>
        /// Displays the progress information of the burning library.
        /// </summary>
        private void m_AudioImage_OnProgress(int Percent, int TimeRemaining)
        {
            if (m_CurrentTrack != null)
            {
                m_CurrentTrack.Progress = (short)Percent;
                olvProgress.RefreshObject(m_CurrentTrack);
            }
        }

        /// <summary>
        /// Displays the progress information of the audio converter.
        /// </summary>
        private void AudioConverter_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            if (m_CurrentTrack != null)
            {
                m_CurrentTrack.Progress = (short)e.ProgressPercentage;
                olvProgress.RefreshObject(m_CurrentTrack);
            }
        }

        protected virtual void OnWorkerDoWork(object sender, DoWorkEventArgs e)
        {
            // If all is identical, do not bother updating (retry situation)
            if (m_WorkItems.Count == m_AudioImage.Tracks.Count)
            {
                bool identical = true;

                for (int i = 0; i < m_AudioImage.Tracks.Count; i++)
                {
                    if (string.Compare(m_AudioImage.Tracks[i].Path, m_WorkItems[i].FileName, true) != 0)
                    {
                        identical = false;
                        break;
                    }
                }

                if (identical)
                {
                    // No work to do
                    UpdateCDTextInformation();
                    return;
                }
            }

            m_AudioImage.Tracks.Clear();

            foreach (ProgressAddingTrack workItem in m_WorkItems)
            {
                if (m_Worker.CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }

                m_CurrentTrack = workItem;
                workItem.Status = Status.Working;
                this.BeginInvoke(new MethodInvoker(delegate() {olvProgress.RefreshObject(workItem);}));
                AudioTrack resultTrack = null;

                string ext = Path.GetExtension(workItem.FileName).ToLower();
                string filename = workItem.FileName;

                // If preprocessing is required, do it now.
                if (AudioConverter.ProcessReplayGain != ReplayGainOption.None && AudioConverter.HasReplayGainInfo(filename))
                {
                    filename = AudioConverter.ToWav(filename, 44100);
                }

                bool trackAdded = false;
                // Analysing MP2 files takes a while, but they are
                // not compatible anyway.
                if (ext != ".mp2")
                {
                    try
                    {
                        resultTrack = m_AudioImage.AddFile(filename);
                        trackAdded = true;
                    }
                    catch (COMException ex)
                    {
                        Debug.LogError(ex);
                    }
                }

                // Unsupported format, try converting
                if (!trackAdded)
                {
                    string altFile = AudioConverter.ToWav(filename, 44100);
                    if (altFile != null)
                    {
                        resultTrack = m_AudioImage.AddFile(altFile);
                    }
                }

                SetCDText(resultTrack, workItem.AudioTrack);

                workItem.Progress = 100;
                workItem.Status = (resultTrack != null) ? Status.Success : Status.Failure;

                this.BeginInvoke(new MethodInvoker(delegate() { olvProgress.RefreshObject(workItem); olvProgress.EnsureVisible(olvProgress.IndexOf(workItem)); }));
            }

            UpdateCDTextInformation();

            if (m_Worker.CancellationPending)
            {
                e.Cancel = true;
            }
        }

        /// <summary>
        /// Updates the CD-Text information of all tracks and the disc.
        /// </summary>
        private void UpdateCDTextInformation()
        {
            m_AudioImage.Artist = CDBurnerXP.Controls.AudioTrack.StripToCDTextSize(m_Layout.CDTextArtist);
            m_AudioImage.Title = CDBurnerXP.Controls.AudioTrack.StripToCDTextSize(m_Layout.CDTextTitle);

            if (m_AudioImage.Tracks.Count == m_Layout.AudioTracks.Length)
            {
                for (int i = 0; i < m_AudioImage.Tracks.Count; i++)
                {
                    SetCDText(m_AudioImage.Tracks[i], m_Layout.AudioTracks[i]);
                }
            }
        }

        /// <summary>
        /// Updates the CD-Text information for a single track.
        /// If no information is given by the user, ID3 tag information (and similar) will be used.
        /// </summary>
        private void SetCDText(AudioTrack resultTrack, CDBurnerXP.Controls.AudioTrack audioTrack)
        {
            if (resultTrack == null)
            {
                return;
            }

            if (string.IsNullOrEmpty(audioTrack.CDTextArtist) && ApplicationSettings.AudioAutoCDText)
            {
                resultTrack.Artist = CDBurnerXP.Controls.AudioTrack.StripToCDTextSize(audioTrack.Artist);
            }
            else
            {
                resultTrack.Artist = CDBurnerXP.Controls.AudioTrack.StripToCDTextSize(audioTrack.CDTextArtist);
            }

            if (string.IsNullOrEmpty(audioTrack.CDTextTitle) && ApplicationSettings.AudioAutoCDText)
            {
                resultTrack.Title = CDBurnerXP.Controls.AudioTrack.StripToCDTextSize(audioTrack.Title);
            }
            else
            {
                resultTrack.Title = CDBurnerXP.Controls.AudioTrack.StripToCDTextSize(audioTrack.CDTextTitle);
            }
        }

        private void m_Worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Cancelled)
            {
                DialogResult = DialogResult.Cancel;
            }
            else
            {
                DialogResult = DialogResult.OK;
            }
            this.Close();
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            if (m_Worker.IsBusy)
            {
                bCancel.Enabled = false;
                m_Worker.CancelAsync();
                e.Cancel = true;
            }
            else if (DialogResult != DialogResult.Cancel)
            {
                int diff = m_Tracks.Length - m_AudioImage.Tracks.Count;
                // Hm, if all tracks failed to add we need to tell the user
                if (m_AudioImage.Tracks.Count == 0)
                {
                    MessageBox.Show(RM.GetString("None of the tracks could be added to the audio image. The disc cannot be burned."), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    DialogResult = DialogResult.Cancel;
                }
                else if (diff > 0)
                {
                    DialogResult = DialogResult.OK;
                    MessageBox.Show(string.Format(RM.GetString("{0} track(s) could not be added to the audio image."), diff), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                }
                else
                {
                    DialogResult = DialogResult.OK;
                }
            }

            base.OnClosing(e);
        }

        protected override void OnClosed(EventArgs e)
        {
            base.OnClosed(e);
            AudioConverter.ProgressChanged -= new EventHandler<ProgressChangedEventArgs>(AudioConverter_ProgressChanged);
            if (m_AudioImage != null)
            {
                m_AudioImage.OnProgress += new IAudioBurnerEvents_OnProgressEventHandler(m_AudioImage_OnProgress);
            }
        }
    }
}