﻿using System;
using CDBurnerXP.Burning;
using CDBurnerXP.Controls;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using NMSDVDXLib;
using RocketDivision.StarBurnX;
using CDBurnerXP.Net;
using System.Runtime.InteropServices;
using Microsoft.Win32;

namespace CDBurnerXP.Forms
{
    public partial class BurningBaseForm : BaseForm
    {
        protected DateTime m_ElapsedTime = DateTime.MinValue;
        protected DateTime m_StartTime = DateTime.MinValue;
        protected int m_CopiesMade = 0;
        private bool m_LoadNumberOfCopiesFromSettings = true;
        protected Status m_Status = Status.Idle;
        protected bool m_BurningProcessCancelled = false;

        /// <summary>
        /// Specifies the current status of the dialog.
        /// </summary>
        public enum Status
        {
            Idle,
            Burning,
            Verifying,
            ReadingFromSource
        }

        #region Properties

        /// <summary>
        /// Gets or sets if the RemainingCalculator class
        /// should be used automatically.
        /// </summary>
        protected virtual bool UseRemainingCalculator
        {
            get { return true; }
        }

        /// <summary>
        /// Gets the currently selected device.
        /// </summary>
        protected virtual DeviceX Device
        {
            get
            {
                return null;
            }
        }

        /// <summary>
        /// Gets the currently selected StarBurn device.
        /// </summary>
        protected virtual Drive StarBurnDevice
        {
            get
            {
                return null;
            }
        }

        /// <summary>
        /// Gets or sets the number of copies a user wants to create.
        /// </summary>
        public virtual int NumberOfCopies
        {
            get
            {
                return 1;
            }
            set
            {
                m_LoadNumberOfCopiesFromSettings = false;
            }
        }

        #endregion

        public BurningBaseForm() : base()
        {
            InitializeComponent();

            AcceptButton = cmdStartBurn;
            CancelButton = cmdCancel;
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            if (m_Status != Status.Idle)
            {
                e.Cancel = true;
                return;
            }

            base.OnClosing(e);

            Settings.SetValue(this, "NumberOfCopies", NumberOfCopies);
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            // Fix bug: BackColor lost after minimising
            if (tabBurnTabs != null)
            {
                tabBurnTabs.Refresh();
            }
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            if (m_LoadNumberOfCopiesFromSettings)
            {
                NumberOfCopies = Conversion.ToInt(Settings.GetValue(this, "NumberOfCopies", NumberOfCopies));
            }
        }

        protected virtual void OnStartBurnClick(object sender, EventArgs e)
        {

        }

        protected virtual void OnCancelClick(object sender, EventArgs e)
        {

        }

        /// <summary>
        /// Performs the necessary actions for the OnCloseSessionBegin event (StarBurn)
        /// </summary>
        protected virtual void HandleCloseSession()
        {
            if (this.InvokeRequired)
            {
                this.Invoke((MethodInvoker)this.HandleCloseSession);
                return;
            }

            if (Animation.FirstFrame != 35)
            {
                Animation.FirstFrame = 35;
                Animation.LastFrame = 49;
                Animation.Delay = 90;
                Animation.Pause = 180;
                Animation.StartAnimation();
            }

            prgProgress.ShowMarquee = true;
            lblStatus.Text = RM.GetString("Closing disc") + "...";
            lvwStatus.AddItem(LogListView.LogType.Information, RM.GetString("Closing disc"));
        }

        /// <summary>
        /// Performs the necessary actions for the start of the burning process.
        /// </summary>
        protected virtual void HandleBurnProcessStarted()
        {
            if (this.InvokeRequired)
            {
                this.Invoke((MethodInvoker)this.HandleBurnProcessStarted);
                return;
            }

            if (this.StarBurnDevice != null)
            {
                BurningEngine.TryLock(this.StarBurnDevice);
            }

            m_BurningProcessCancelled = false;

            if (Animation.FirstFrame != 124)
            {
                Animation.FirstFrame = 124;
                Animation.LastFrame = 143;
                Animation.Delay = 90;
                Animation.Pause = 180;
                Animation.StartAnimation();
            }

            prgProgress.Value = 0;
            SetBurnProcessStartedText();
        }

        protected virtual void SetBurnProcessStartedText()
        {
            lvwStatus.AddItem(LogListView.LogType.Information, RM.GetString("Burning process started."));
            lblStatus.Text = RM.GetString("Preparing to write disc...");
        }

        /// <summary>
        /// Performs the necessary actions for the event that an error occured.
        /// </summary>
        protected virtual void HandleError(object ex)
        {
            HandleError(ex, false, false, STARBURN_RESULT_CODES.STARBURN_OPERATION_SUCCEEDED);
        }

        protected virtual void HandleTestWriteDisabled(ref bool cancelBurning)
        {
            if (MessageBox.Show(this, RM.GetString("The current disc does not support simulation of the burning process. Do you want to skip the simulation and perform the real burning process?"), RM.GetString("Simulation not supported"), MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2) == DialogResult.No)
            {
                m_BurningProcessCancelled = true;
                lvwStatus.AddItem(LogListView.LogType.Information, RM.GetString("The writing was cancelled by the user, aborting please wait.."));
                cancelBurning = true;
            }
        }

        /// <summary>
        /// Performs the necessary actions for the event that an error occured.
        /// </summary>
        protected virtual void HandleError(object ex, bool loggingEnabled, bool shutdown, STARBURN_RESULT_CODES errorCode)
        {
            if (shutdown)
            {
                ExitWindows.ForceShutdown();
            }

            tmrElapsed.Enabled = false;

            if (Animation.FirstFrame != 62)
            {
                Animation.FirstFrame = 62;
                Animation.LastFrame = 73;
                Animation.Delay = 90;
                Animation.Pause = 500;
                Animation.StartAnimation();
            }

            prgProgress.ShowMarquee = false;
            prgProgress.Value = 0;

            COMException comException = ex as COMException;

            // Show different error message if only a verification error happened
            if (errorCode == STARBURN_RESULT_CODES.STARBURN_ERR_VERIFY_DISC_FAILED)
            {
                lvwStatus.AddItem(LogListView.LogType.Error, RM.GetString("Error occured:") + " " + RM.GetString("Verification failed"));
                MessageBox.Show(this, RM.GetString("The burning process has been completed, but the verification of the burned data failed. The data on the disc might be unusable."), RM.GetString("Verification failed"), MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            else if (comException.IsError(STARBURN_RESULT_CODES.STARBURN_ERR_DISC_NO_FREE_SPACE))
            {
                lvwStatus.AddItem(LogListView.LogType.Error, RM.GetString("Error occured:") + " " + RM.GetString("Insufficient free space on disc"));
                MessageBox.Show(this, RM.GetString("There is not enough free space on the disc. Please free up some disc space or choose a different target."), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                lvwStatus.AddItem(LogListView.LogType.Error, RM.GetString("Error occured:") + " " + comException.Message);

                BurnErrorReport errorReport = new BurnErrorReport(comException);
                errorReport.ShowErrorAndAskForReport(this, loggingEnabled);
            }

            ResetUiAfterWork();

#if !DEBUG
            if (StarBurnDevice != null)
            {
                StarBurnDevice.Eject();
            }
#endif
        }

        protected virtual void ActivateControls(bool enable)
        {
        }

        protected virtual void PrepareUiForWork()
        {
            cmdCancel.Text = RM.GetString("Cancel");
            cmdStartBurn.Enabled = false;
            tabBurnTabs.SelectedTab = tabBurnProgress;
            ActivateControls(false);
            // Stop Notifying While Burning
            MediaNotifier.Enabled = false;

            // make sure, that all previous media notifier requests are processed
            Application.DoEvents();
        }

        protected virtual void ResetUiAfterWork()
        {
            if (this.StarBurnDevice != null)
            {
                try
                {
                    this.StarBurnDevice.UnlockTray();
                }
                catch (COMException)
                {
                    // Ignore errors
                }
            }

            ActivateControls(true);
            cmdCancel.Tag = null;
            cmdCancel.Text = RM.GetString("Close");
            tabBurnTabs.SelectedTab = tabBurnOptions;
            cmdStartBurn.Enabled = true;
            Animation.ShowFrame(0);
            tmrElapsed.Stop();
            lblStatus.Text = RM.GetString("Status: Idle");
            lblTime.Text = String.Format(RM.GetString("Time Elapsed (hh:mm:ss): {0}"), "");
            lblRemaining.Text = String.Format(RM.GetString("Time Remaining (hh:mm:ss): {0}"), "");
            prgProgress.ShowMarquee = false;

            // Start Notifying again
            MediaNotifier.Enabled = true;
            m_Status = Status.Idle;
        }

        /// <summary>
        /// Determines whether or not further copies need to be burned.
        /// </summary>
        protected virtual bool FurtherCopiesRequired()
        {
            m_CopiesMade += 1;
            if (NumberOfCopies > m_CopiesMade)
            {
                return true;
            }

            // Start over
            m_CopiesMade = 0;
            return false;
        }

        /// <summary>
        /// Ejects the disc, forcing the user to
        /// insert a new disc, then starts a new burning process
        /// </summary>
        protected virtual void BeginBurnNextDisc()
        {
            Timer timer = new Timer();
            timer.Interval = 100;
            timer.Tick += new EventHandler(timer_Tick);
            timer.Start();
        }

        protected void TryEject(Drive drive)
        {
            try
            {
                drive.Eject();
            }
            catch (COMException e)
            {
                lvwStatus.AddItem(LogListView.LogType.Error, e.Message);
            }
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            ((Timer)sender).Stop();

            try
            {
                this.BeginInvoke((MethodInvoker)delegate()
                {
                    if (Device != null)
                    {
                        Device.Eject();
                    }
                    else
                    {
                        StarBurnDevice.Eject();
                    }

                    OnStartBurnClick(this, null);
                });
            }
            catch (InvalidOperationException)
            {
            }
        }


        private void tmrElapsed_Tick(object sender, EventArgs e)
        {
            m_ElapsedTime = DateTime.Now;
            TimeSpan deltaTime = Utility.TimeSpanSeconds(m_ElapsedTime - m_StartTime);
            lblTime.Text = String.Format(RM.GetString("Time Elapsed (hh:mm:ss): {0}"), deltaTime);

            try
            {
                // Make sure that system does not enter standby mode when burning (especially large discs)
                Kernel32.SetThreadExecutionState(Kernel32.EXECUTION_STATE.ES_SYSTEM_REQUIRED);
            }
            catch (Exception)
            {
                // Not critical
            }

            if (UseRemainingCalculator)
            {
                lblRemaining.Text = String.Format(RM.GetString("Time Remaining (hh:mm:ss): {0}"), RemainingCalculator.GetRemainingTime(deltaTime, prgProgress.Value));
            }
        }

        protected virtual void SetTimeRemaining(int seconds)
        {
            this.SetTimeRemaining(TimeSpan.FromSeconds(seconds));
        }

        protected virtual void SetTimeRemaining(TimeSpan timeRemaining)
        {
            try
            {
                // Make sure that system does not enter standby mode when burning (especially large discs)
                Kernel32.SetThreadExecutionState(Kernel32.EXECUTION_STATE.ES_SYSTEM_REQUIRED);
            }
            catch (Exception)
            {
                // Not critical
            }

            if (timeRemaining.TotalSeconds <= 60)
            {
                lblRemaining.Text = String.Format(RM.GetString("Time Remaining (hh:mm:ss): {0}", Culture), RM.GetString("< 1 minute", Culture));
            }
            else
            {
                // Only use approximate duration, not very precise anyway.
                int minutes = Conversion.ToInt(Math.Ceiling(timeRemaining.TotalMinutes));
                lblRemaining.Text = FormatTranslation("Time Remaining (hh:mm:ss): {0}", FormatTranslation("{0:0} minutes", minutes));
            }
        }
    }
}
