﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using Microsoft.Win32;
using CDBurnerXP.IO;

namespace CDBurnerXP.Forms
{
    public partial class DiskSpaceRequiredDialog : BaseForm
    {
        private long m_NumberOfFreeBytesRequired = 0;
        private string m_LocationToCheck = null;
        private const long fatMaxFileSize = 4294967295; // (2^32)-1

        #region Properties

        protected override bool AutoTranslate
        {
            get
            {
                return true;
            }
        }

        protected long FreeBytesRequired
        {
            get
            {
                return m_NumberOfFreeBytesRequired;
            }
            set
            {
                m_NumberOfFreeBytesRequired = value;
            }
        }

        protected string TargetLocation
        {
            set
            {
                m_LocationToCheck = value;
            }
            get
            {
                if (m_LocationToCheck != null)
                {
                    return m_LocationToCheck;
                }
                return PathEx.GetTempPath();
            }
        }

        protected long FreeDiskSpace
        {
            get
            {
                long lpFreeBytesAvailable = 0;
                long lpTotalNumberOfBytes = 0;
                long lpTotalNumberOfFreeBytes = 0;
                Kernel32.GetDiskFreeSpaceEx(Kernel32.GetVolumePathName(TargetLocation), ref lpFreeBytesAvailable, ref lpTotalNumberOfBytes, ref lpTotalNumberOfFreeBytes);
                return lpFreeBytesAvailable;
            }
        }

        /// <summary>
        /// Requires an addional 10% of free bytes, 50MiB at least.
        /// </summary>
        protected long TotalFreeBytesRequired
        {
            get
            {
                long tenPercent = Convert.ToInt64(m_NumberOfFreeBytesRequired * 0.1);
                int fiftyMib = 50 * 1024 * 1024;

                if (tenPercent < fiftyMib)
                {
                    return fiftyMib;
                }
                return FreeBytesRequired + tenPercent;
            }
        }

        protected bool EnoughFreeSpaceAvailable
        {
            get
            {
                if (FreeDiskSpace <= 0)
                {
                    // Disc space could not be determined - assume there is enough available,
                    // so that our inability does to block the whole operation.
                    return true;
                }
                return (FreeDiskSpace >= TotalFreeBytesRequired);
            }
        }

        #endregion

        private DiskSpaceRequiredDialog()
        {
            InitializeComponent();
            CancelButton = bCancel;
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            lblValueAvailable.Text = FormatFileSize.Format(FreeDiskSpace);
            lblValueRequired.Text = FormatFileSize.Format(TotalFreeBytesRequired);
            lblValueLocation.Text = Utility.CompactString(TargetLocation, tblInformation.Width - lblLocation.Width - 15, Font, "");
        }

        private void tmrRefresh_Tick(object sender, EventArgs e)
        {
            long freeDiskSpace = FreeDiskSpace;
            lblValueAvailable.Text = FormatFileSize.Format(freeDiskSpace);

            if (EnoughFreeSpaceAvailable)
            {
                DialogResult = DialogResult.OK;
                Close();
            }
        }

        /// <summary>
        /// Shows the disk space dialog and returns true if
        /// enough disk space has been aquired.
        /// </summary>
        /// <param name="owner">Owner window of the dialog</param>
        /// <param name="bytesRequired">How many bytes are required to proceed.
        /// The dialog will demand an addional 10% (at least 50MiB) to ensure system stability.</param>
        /// <returns>true if enough disk space has been aquired, false otherwise</returns>
        public static bool Show(IWin32Window owner, long bytesRequired)
        {
            return Show(owner, bytesRequired, null);
        }

        /// <summary>
        /// Shows the disk space dialog and returns true if
        /// enough disk space has been aquired.
        /// </summary>
        /// <param name="owner">Owner window of the dialog</param>
        /// <param name="bytesRequired">How many bytes are required to proceed.
        /// The dialog will demand an addional 10% (at least 50MiB) to ensure system stability.</param>
        /// <returns>true if enough disk space has been aquired, false otherwise</returns>
        public static bool Show(IWin32Window owner, long bytesRequired, string location)
        {
            DiskSpaceRequiredDialog dialog = new DiskSpaceRequiredDialog();
            dialog.FreeBytesRequired = bytesRequired;
            dialog.TargetLocation = location;

            // Check if target file system is FAT and the file exteeds 4GB
            if (bytesRequired > fatMaxFileSize)
            {
                string volumeRoot = Kernel32.GetVolumePathName(location);
                foreach (DriveInfo info in DriveInfo.GetDrives())
                {
                    if (string.Compare(info.Name, volumeRoot, true) == 0)
                    {
                        if (info.DriveFormat == "FAT32")
                        {
                            MessageBox.Show(owner, string.Format(RM.GetString("The data cannot be saved to drive {0}, because its file system is FAT32 and cannot hold any files greater than 4GB ({1} bytes).\r\n\r\nIn order to proceed, use a different drive, which uses the NTFS file system, as target."), info.Name, fatMaxFileSize), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                            return false;
                        }
                    }
                }
            }

            // Don't show dialog if not necessary in the first place
            if (dialog.EnoughFreeSpaceAvailable) return true;

            try
            {
                return (dialog.ShowDialog(owner) != DialogResult.Cancel);
            }
            finally
            {
                dialog.Dispose();
            }
        }
    }
}
