using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Microsoft.Win32
{
    public sealed class ExitWindows
    {
        /// <remarks />
        private ExitWindows()
        {
        }

        #region "Public Methods"
        /// <summary>
        /// Properly adjusts the shutdown privilege for this process and shuts down Windows.
        /// </summary>
        /// <param name="Method">The method to use for shutdown.</param>
        /// <returns>True on success, false on error.</returns>
        public static bool Shutdown(ShutdownMethod Method)
        {
            // Adjust the process privilege
            bool ret = AdjustShutdownTokenPriveleges();
            if (!ret)
            {
                return false;
            }

            // exit windows
            return ExitWindowsEx(Method, ShutdownReason.FlagPlanned);
        }


        /// <summary>
        /// Properly adjusts the shutdown privilege for this process and shuts down Windows.
        /// </summary>
        /// <param name="Method">The method to use for shutdown.</param>
        /// <param name="Reason">The reason for the shutdown action.</param>
        /// <returns>True on success, false on error.</returns>
        public static bool Shutdown(ShutdownMethod Method, ShutdownReason Reason)
        {
            // Adjust the process privilege
            bool ret = AdjustShutdownTokenPriveleges();
            if (!ret)
            {
                return false;
            }

            // exit windows
            return ExitWindowsEx(Method, Reason);
        }

        /// <summary>
        /// Tries to hibernate the computer, or forces a shutdown if that is not possible.
        /// </summary>
        /// <returns></returns>
        public static bool HibernateOrShutdown()
        {
            try
            {
                Application.SetSuspendState(PowerState.Hibernate, true, true);
                return true;
            }
            catch (Exception)
            {
                return ForceShutdown();
            }
        }

        /// <summary>
        /// Forces the computer to shut down.
        /// </summary>
        public static bool ForceShutdown()
        {
            return ExitWindows.Shutdown(ExitWindows.ShutdownMethod.PowerOff | ExitWindows.ShutdownMethod.Force);
        }

        #endregion

        #region "Private Methods"
        /// <summary>
        /// Adjust the token of the current process to include shutdown privileges.
        /// </summary>
        /// <returns>True on success, false on error.</returns>
        private static bool AdjustShutdownTokenPriveleges()
        {
            bool ret;

            IntPtr hProc = IntPtr.Zero;
            IntPtr hToken = IntPtr.Zero;
            LUID luidRestore;
            TOKEN_PRIVILEGES tokenPriviliges;

            // get the current current process security token
            hProc = System.Diagnostics.Process.GetCurrentProcess().Handle;
            ret = OpenProcessToken(hProc, TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, ref hToken);
            if (!ret)
            {
                return false;
            }


            // lookup the LUID for the shutdown privilege
            ret = LookupPrivilegeValue(String.Empty, SE_SHUTDOWN_NAME, out luidRestore);
            if (!ret)
            {
                return false;
            }

            // adjust the privileges of the current process to include the shutdown privilege
            tokenPriviliges.PrivilegeCount = 1;
            tokenPriviliges.Luid = luidRestore;
            tokenPriviliges.Attributes = SE_PRIVILEGE_ENABLED;

            ret = AdjustTokenPrivileges(hToken, false, ref tokenPriviliges, 0, IntPtr.Zero, IntPtr.Zero);
            if (!ret)
            {
                return false;
            }

            return true;
        }
        #endregion

        #region "Shutdown argument type definitions"
        /// <summary>
        /// Defines the methods of shutting down the system.
        /// </summary>
        [Flags()]
        public enum ShutdownMethod : uint
        {
            LogOff = 0,
            ShutDown = 1,
            Reboot = 2,
            Force = 4,
            PowerOff = 8,
            ForceIfHung = 16
        }

        /// <summary>
        /// Defines the reason for the shutdown of the system.
        /// </summary>
        [Flags()]
        public enum ShutdownReason : uint
        {
            MajorApplication = 262144,
            MajorHardware = 65536,
            MajorLegacyApi = 458752,
            MajorOperatingSystem = 131072,
            MajorOther = 0,
            MajorPower = 393216,
            MajorSoftware = 196608,
            MajorSystem = 327680,

            MinorBlueScreen = 15,
            MinorCordUnplugged = 11,
            MinorDisk = 7,
            MinorEnvironment = 12,
            MinorHardwareDriver = 13,
            MinorHotfix = 17,
            MinorHung = 5,
            MinorInstallation = 2,
            MinorMaintenance = 1,
            MinorMMC = 25,
            MinorNetworkConnectivity = 20,
            MinorNetworkCard = 9,
            MinorOther = 0,
            MinorOtherDriver = 14,
            MinorPowerSupply = 10,
            MinorProcessor = 8,
            MinorReconfig = 4,
            MinorSecurity = 19,
            MinorSecurityFix = 18,
            MinorSecurityFixUninstall = 24,
            MinorServicePack = 16,
            MinorServicePackUninstall = 22,
            MinorTermSrv = 32,
            MinorUnstable = 6,
            MinorUpgrade = 3,
            MinorWMI = 21,

            FlagUserDefined = 1073741824,
            FlagPlanned = 2147483648
        }

        #endregion

        #region "Win32 Imports"
        [DllImport("user32.dll")]
        private static extern bool ExitWindowsEx(ShutdownMethod uMethod, ShutdownReason dwReason);

        [DllImport("advapi32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool AdjustTokenPrivileges(
           IntPtr TokenHandle,
           [MarshalAs(UnmanagedType.Bool)]bool DisableAllPrivileges,
           ref TOKEN_PRIVILEGES NewState,
           UInt32 Zero,
           IntPtr Null1,
           IntPtr Null2);

        [DllImport("advapi32.dll", SetLastError = true)]
        private static extern bool OpenProcessToken(IntPtr ProcessHandle, UInt32 DesiredAccess, ref IntPtr TokenHandle);

        [DllImport("advapi32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        private static extern bool LookupPrivilegeValue(string lpSystemName, string lpName, out LUID lpLuid);

        [StructLayout(LayoutKind.Sequential)]
        private struct TOKEN_PRIVILEGES
        {
            public int PrivilegeCount;
            public LUID Luid;
            public int Attributes;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct LUID
        {
            public uint LowPart;
            public uint HighPart;
        }

        private const int TOKEN_QUERY = 8;
        private const int TOKEN_ADJUST_PRIVILEGES = 32;
        private const string SE_SHUTDOWN_NAME = "SeShutdownPrivilege";
        private const int SE_PRIVILEGE_ENABLED = 2;

        #endregion
    }
}
